"""Feature set."""
from typing import Any, Callable, Dict

import haiku as hk
import jax.numpy as jnp
import tree

from moss.network.feature.feature import BaseFeature
from moss.types import Array


class FeatureSet(object):
  """Feature set."""

  def __init__(
    self, name: str, features: Dict[str, BaseFeature],
    encoder_net_maker: Callable[[], hk.Module]
  ) -> None:
    """Init."""
    self._name = name
    self._features = features
    self._encoder_net_maker = encoder_net_maker

  @property
  def name(self) -> str:
    """Get name."""
    return self._name

  @property
  def spec(self) -> Dict[str, BaseFeature]:
    """Get features spec dict."""
    return self._features

  @property
  def encoder_net_maker(self) -> Callable[[], Any]:
    """Feature encoder."""
    return self._encoder_net_maker

  def generate_value(self) -> Any:
    """Generate a test value which conforms to this feature set."""
    return tree.map_structure(lambda spec: spec.generate_value(), self.spec)

  def process(self, feature_inputs: Dict) -> Array:
    """Feature inputs process."""
    output_list = []
    for name, feature in self._features.items():
      inputs = feature_inputs.get(name)
      if inputs is None:
        raise ValueError(
          f"Not find the inputs named `{name}` in `{self.name}` feature set, "
          "please check you inputs."
        )
      output_list.append(feature.process(inputs))
    output = jnp.concatenate(output_list, axis=0)
    return output



"""Torso network."""
from moss.network.torso.dense import DenseTorso
from moss.network.torso.gru import GRUTorso
from moss.network.torso.lstm import LSTMTorso

__all__ = [
  "DenseTorso",
  "LSTMTorso",
  "GRUTorso",
]



"""Atari local run."""
import os
import pickle
from functools import partial

import jax
import numpy as np
import tree
from absl import app, flags, logging
from dm_env import TimeStep

from examples.atari.agent import AtariAgent
from examples.atari.network import network_maker
from examples.atari.utils import LocalEnv
from moss.predictor import BasePredictor
from moss.types import Params
from moss.utils.loggers import TerminalLogger

flags.DEFINE_string("task_id", "Pong-v5", "Task name.")
flags.DEFINE_string("model_path", None, "Restore model path.")

FLAGS = flags.FLAGS


def main(_):
  """Main."""
  local_env = LocalEnv(FLAGS.task_id, scale=2)
  obs_spec = local_env.observation_spec()
  action_spec = local_env.action_spec()

  predictor = BasePredictor(
    1, partial(network_maker, obs_spec, action_spec), TerminalLogger
  )
  with open(FLAGS.model_path, mode="rb") as f:
    params: Params = pickle.load(f)
    predictor.update_params(params)
  agent = AtariAgent(0, None, predictor, TerminalLogger())

  rng = jax.random.PRNGKey(42)
  total_reward = 0
  timestep: TimeStep = local_env.reset()
  while True:
    if timestep.first():
      total_reward = 0
      agent.reset()

    input_dict, reward = agent.step(timestep)
    total_reward += reward
    sub_key, rng = jax.random.split(rng)
    input_dict = tree.map_structure(lambda x: np.expand_dims(x, 0), input_dict)
    action, _ = predictor._forward(params, input_dict, None, sub_key)
    take_action = agent.take_action(action)
    take_action = np.array(take_action)
    timestep = local_env.step(take_action)

    if timestep.last():
      _, reward = agent.step(timestep)
      total_reward += reward
      logging.info(f"Total reward: {total_reward}")


if __name__ == "__main__":
  os.environ["CUDA_VISIBLE_DEVICES"] = ""
  app.run(main)



"""PettingZoo network."""
from functools import partial
from typing import Any

import numpy as np

from moss.network import CommonNet
from moss.network.action import ActionSpec, DiscreteAction
from moss.network.feature import (
  FeatureSet,
  FeatureSpec,
  ImageFeature,
  OneHotFeature,
)
from moss.network.feature.encoder import (
  CommonEncoder,
  Conv2DConfig,
  ImageFeatureEncoder,
  ResnetConfig,
)
from moss.network.torso import DenseTorso
from moss.network.value import DenseValue

resnet_default_config = [
  ResnetConfig(16, 2),
  ResnetConfig(16, 2),
  ResnetConfig(32, 2),
]
conv2d_default_config = [
  Conv2DConfig(32, 8, 4, "VALID"),
  Conv2DConfig(64, 4, 2, "VALID"),
  Conv2DConfig(64, 3, 1, "VALID"),
]


def network_maker(
  obs_spec: Any,
  action_spec: Any,
  data_format: str = "NHWC",
  use_resnet: bool = False,
  use_orthogonal: bool = True,
) -> Any:
  """Atari network maker."""
  height, width, channel = obs_spec["first_0"].shape
  num_actions = action_spec["first_0"].n

  pettingzoo_frame = FeatureSet(
    name="pettingzoo_frame",
    features={
      "frame":
        ImageFeature(
          height, width, channel, data_format, np.int8, "frame",
          lambda x: x / 255.
        )
    },
    encoder_net_maker=lambda: ImageFeatureEncoder(
      "frame_encoder",
      data_format,
      use_resnet=use_resnet,
      resnet_config=resnet_default_config,
      conv2d_config=conv2d_default_config,
      use_orthogonal=use_orthogonal
    )
  )
  player = FeatureSet(
    name="player",
    features={
      "id": OneHotFeature(2, name="id"),
    },
    encoder_net_maker=lambda: CommonEncoder(
      "player_encoder",
      hidden_sizes=[32],
      use_orthogonal=use_orthogonal,
    )
  )

  feature_sets = {
    "pettingzoo_frame": pettingzoo_frame,
    "player": player,
  }
  actions = {
    "pettingzoo_action":
      DiscreteAction("pettingzoo_action", [512], num_actions, use_orthogonal),
  }

  torso_net_maker = partial(DenseTorso, "torso", [512], use_orthogonal)
  value_net_maker = partial(DenseValue, "value", [512, 32], use_orthogonal)
  return CommonNet(
    feature_spec=FeatureSpec(feature_sets),
    action_spec=ActionSpec(actions),
    torso_net_maker=torso_net_maker,
    value_net_maker=value_net_maker,
  )



"""Feature network."""
from moss.network.feature.feature import (
  BaseFeature,
  ImageFeature,
  OneHotFeature,
  ScalarFeature,
  VectorFeature,
)
from moss.network.feature.feature_set import FeatureSet
from moss.network.feature.feature_spec import FeatureSpec

__all__ = [
  "BaseFeature",
  "ImageFeature",
  "OneHotFeature",
  "ScalarFeature",
  "VectorFeature",
  "FeatureSet",
  "FeatureSpec",
]



"""Action network."""
from moss.network.action.action_spec import ActionSpec
from moss.network.action.base import Action
from moss.network.action.discrete import DiscreteAction

__all__ = [
  "ActionSpec",
  "Action",
  "DiscreteAction",
]



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Logger that flattens nested data."""

from typing import Sequence

from moss.utils.loggers import base


class FlattenDictLogger(base.Logger):
  """Logger which flattens sub-dictionaries into the top level dict."""

  def __init__(
    self,
    logger: base.Logger,
    label: str = 'Logs',
    raw_keys: Sequence[str] = ()
  ):
    """Initializer.

    Args:
      logger: The wrapped logger.
      label: The label to add as a prefix to all keys except for raw ones.
      raw_keys: The keys that should not be prefixed. The values for these keys
        must always be flat. Metric visualisation tools may require certain
        keys to be present in the logs (e.g. 'step', 'timestamp'), so these
        keys should not be prefixed.
    """
    self._logger = logger
    self._label = label
    self._raw_keys = raw_keys

  def write(self, values: base.LoggingData) -> None:
    """Write."""
    flattened_values = {}
    for key, value in values.items():
      if key in self._raw_keys:
        flattened_values[key] = value
        continue
      name = f'{self._label}/{key}'
      if isinstance(value, dict):
        for sub_key, sub_value in value.items():
          flattened_values[f'{name}/{sub_key}'] = sub_value
      else:
        flattened_values[name] = value

    self._logger.write(flattened_values)

  def close(self) -> None:
    """Close."""
    self._logger.close()



"""Doom agent."""
from typing import Any, Dict, Tuple

import jax.numpy as jnp

from moss.agent.base import BaseAgent
from moss.core import Buffer, Predictor
from moss.network.keys import AGENT_STATE
from moss.types import LoggingData, Reward, TimeStep
from moss.utils.loggers import Logger


class DoomAgent(BaseAgent):
  """Doom agent."""

  def __init__(
    self, unroll_length: int, buffer: Buffer, predictor: Predictor,
    logger: Logger
  ) -> None:
    """Init."""
    super().__init__("Doom", unroll_length, buffer, predictor, logger)
    self._episode_steps: int = 0
    self._rewards: float = 0

  def _init(self) -> None:
    """Init agent states."""
    self._episode_steps = 0
    self._rewards = 0

  def reset(self) -> LoggingData:
    """Reset agent."""
    metrics = {
      f"{self._name}/episode steps": self._episode_steps,
      f"{self._name}/total rewards": self._rewards
    }
    self._init()
    return metrics

  def step(self, timestep: TimeStep) -> Tuple[Dict, Reward]:
    """Agent step.

    Return:
      input_dict: input dict for network.
        Returns must be serializable Python object to ensure that it can
        exchange data between launchpad's nodes.
    """
    obs = timestep.observation.obs
    obs = jnp.transpose(obs, axes=(1, 2, 0))
    state = {"doom_frame": {"frame": jnp.array(obs)}}
    input_dict = {AGENT_STATE: state}
    reward = timestep.reward
    self._episode_steps += 1
    self._rewards += reward
    return input_dict, reward

  def take_action(self, action: Dict[str, Any]) -> Any:
    """Take action."""
    return action["doom_action"]



"""Atari local run."""
import os
import pickle
from functools import partial
from typing import Any, Dict

import jax
import numpy as np
import tree
from absl import app, flags, logging
from dm_env import TimeStep

from examples.pettingzoo.agent import PettingZooAgent
from examples.pettingzoo.network import network_maker
from moss.env.pettingzoo import PettingZooEnv
from moss.predictor import BasePredictor
from moss.types import Params
from moss.utils.loggers import TerminalLogger

flags.DEFINE_string("task_id", "pong_v3", "Task name.")
flags.DEFINE_string("model_path", None, "Restore model path.")

FLAGS = flags.FLAGS


def main(_):
  """Main."""
  local_env = PettingZooEnv(FLAGS.task_id, render_mode="human")
  obs_spec = local_env.observation_spec()
  action_spec = local_env.action_spec()

  predictor = BasePredictor(
    1, partial(network_maker, obs_spec, action_spec), TerminalLogger
  )
  with open(FLAGS.model_path, mode="rb") as f:
    params: Params = pickle.load(f)
    predictor.update_params(params)
  agent = PettingZooAgent(0, None, predictor, TerminalLogger())

  rng = jax.random.PRNGKey(42)
  timestep_dict: Dict[Any, TimeStep] = local_env.reset()
  rewards, actions = {}, {}
  while True:
    for agent_id, timestep in timestep_dict.items():
      if timestep.first():
        rewards[agent_id] = 0
        agent.reset()

      input_dict, reward = agent.step(timestep)
      rewards[agent_id] += reward
      sub_key, rng = jax.random.split(rng)
      input_dict = tree.map_structure(lambda x: np.expand_dims(x, 0), input_dict)
      action, _ = predictor._forward(params, input_dict, None, sub_key)
      take_action = agent.take_action(action)
      actions[agent_id] = int(take_action[0])

      if timestep.last():
        _, reward = agent.step(timestep)
        rewards[agent_id] += reward
        logging.info(f"{agent_id} total reward: {rewards[agent_id]}")
    timestep_dict = local_env.step(actions)


if __name__ == "__main__":
  os.environ["CUDA_VISIBLE_DEVICES"] = ""
  app.run(main)



"""Action spec."""
import collections
from typing import Any, Dict

from distrax import DistributionLike

from moss.network.action.base import Action
from moss.types import Array, KeyArray, SpecArray

ActionDistribution = collections.namedtuple(
  "ActionDistribution", ["log_prob", "prob", "entropy"]
)


class ActionSpec(object):
  """Action spec."""

  def __init__(self, actions: Dict[str, Action]) -> None:
    """Init."""
    self._actions = actions
    self._spec = {name: action.spec for name, action in actions.items()}

  @property
  def actions(self) -> Dict[str, Action]:
    """Actions."""
    return self._actions

  @property
  def action_spec(self) -> Dict[str, SpecArray]:
    """Action specs."""
    return self._spec

  def distribution(self, logits: Dict[str, Any]) -> DistributionLike:
    """Action distribution."""

    def log_prob(value: Dict[str, Array]) -> Array:
      """See `Distribution.log_prob`."""
      prob: Array = 0.0  # type: ignore
      for name, action in self._actions.items():
        prob += action.distribution(logits[name]).log_prob(value[name])
      return prob

    def prob(value: Dict[str, Array]) -> Array:
      """See `Distribution.prob`."""
      prob: Array = 1.0  # type: ignore
      for name, action in self._actions.items():
        prob *= action.distribution(logits[name]).prob(value[name])
      return prob

    def entropy() -> Array:
      """See `Distribution.entropy`."""
      ent: Array = 0.0  # type: ignore
      for name, action in self._actions.items():
        ent += action.distribution(logits[name]).entropy()
      return ent

    return ActionDistribution(log_prob, prob, entropy)

  def sample(self, rng: KeyArray, logits: Dict[str, Any]) -> Any:
    """Sample actions."""
    actions = {
      name: action.sample(rng, logits[name])
      for name, action in self._actions.items()
    }
    return actions



"""Env wrappers."""
from moss.env.wrapper.gym import GymToDeepmindWrapper
from moss.env.wrapper.pettingzoo import PettingZooToDeepmindWrapper

__all__ = [
  "GymToDeepmindWrapper",
  "PettingZooToDeepmindWrapper",
]



# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Residual block."""
from typing import Any, Optional

import haiku as hk
import jax

from moss.types import Array


class ResidualBlock(hk.Module):
  """Residual block."""

  def __init__(
    self,
    num_channels: int,
    name: Optional[str] = None,
    data_format: str = "NHWC",
    use_orthogonal: bool = True
  ) -> None:
    """Init."""
    super().__init__(name=name)
    self._num_channels = num_channels
    self._data_format = data_format
    self._use_orthogonal = use_orthogonal

  def __call__(self, inputs: Array) -> Any:
    """Call."""
    w_init = hk.initializers.Orthogonal() if self._use_orthogonal else None
    main_branch = hk.Sequential(
      [
        jax.nn.relu,
        hk.Conv2D(
          self._num_channels,
          kernel_shape=[3, 3],
          stride=[1, 1],
          w_init=w_init,
          padding="SAME",
          data_format=self._data_format
        ),
        jax.nn.relu,
        hk.Conv2D(
          self._num_channels,
          kernel_shape=[3, 3],
          stride=[1, 1],
          w_init=w_init,
          padding="SAME",
          data_format=self._data_format
        ),
      ]
    )
    return main_branch(inputs) + inputs



"""Dense torso network."""
from typing import Any, Dict, List, Tuple

import haiku as hk
import jax
import jax.numpy as jnp

from moss.types import Array


class DenseTorso(hk.Module):
  """Dense torso network."""

  def __init__(
    self, name: str, hidden_sizes: List[int], use_orthogonal: bool = True
  ):
    """Init."""
    super().__init__(name)
    self._hidden_sizes = hidden_sizes
    self._use_orthogonal = use_orthogonal

  def __call__(self, inputs: Dict[str, Any],
               rnn_state: Any) -> Tuple[Array, Any]:
    """Dense torso network forward.

    Args:
      inputs: input features.
      rnn_state: rnn state, ignore for this class.
    """
    mlp_layers: List[Any] = []
    w_init = hk.initializers.Orthogonal() if self._use_orthogonal else None
    for hidden_size in self._hidden_sizes:
      mlp_layers.append(hk.Linear(hidden_size, w_init=w_init))
      mlp_layers.append(jax.nn.relu)
    torso_net = hk.Sequential(mlp_layers)
    torso_input = jnp.concatenate(list(inputs.values()), axis=-1)
    torso_out = torso_net(torso_input)
    return torso_out, rnn_state



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Default logger."""

import logging
import os
from functools import partial
from typing import Any, Callable, List, Mapping, Optional

from moss.utils.loggers import aggregators
from moss.utils.loggers import asynchronous as async_logger
from moss.utils.loggers import base, csv, filters, tensorboard, terminal


def make_default_logger(
  label: str,
  use_csv: bool = True,
  use_tb: bool = True,
  log_dir: str = "logs",
  time_delta: float = 1.0,
  asynchronous: bool = False,
  print_fn: Optional[Callable[[str], None]] = None,
  serialize_fn: Optional[Callable[[Mapping[str, Any]], str]] = base.to_numpy,
  steps_key: str = 'steps',
) -> base.Logger:
  """Makes a default Moss logger.

  Args:
    label: Name to give to the logger.
    use_csv: Whether use csv logger to persist data.
    use_tb: Whether use tensorboard logger to persist data.
    log_dir: Directory for persist log data.
    time_delta: Time (in seconds) between logging events.
    asynchronous: Whether the write function should block or not.
    print_fn: How to print to terminal (defaults to print).
    serialize_fn: An optional function to apply to the write inputs before
      passing them to the various loggers.
    steps_key: Ignored.

  Returns:
    A logger object that responds to logger.write(some_dict).
  """
  del steps_key
  if not print_fn:
    print_fn = logging.info
  terminal_logger = terminal.TerminalLogger(label=label, print_fn=print_fn)

  loggers: List[base.Logger] = [terminal_logger]

  if use_csv:
    csv_log_dir = os.path.join(log_dir, "csv")
    loggers.append(csv.CSVLogger(csv_log_dir, label, add_uid=False))
  if use_tb:
    tb_log_dir = os.path.join(log_dir, "tb")
    loggers.append(tensorboard.TBXSummaryLogger(tb_log_dir, label))

  # Dispatch to all writers and filter Nones and by time.
  logger = aggregators.Dispatcher(loggers, serialize_fn)
  logger = filters.NoneFilter(logger)
  if asynchronous:
    logger = async_logger.AsyncLogger(logger)  # type: ignore
  logger = aggregators.TimeAggregator(logger, time_delta)

  return logger


def experiment_logger_factory(
  project: str,
  uid: str,
  log_dir: str = "logs",
  **kwargs: Any,
) -> Callable[..., base.Logger]:
  """Experiment logger factory."""
  log_dir = os.path.join(log_dir, project, uid)
  logger_maker = partial(make_default_logger, log_dir=log_dir, **kwargs)
  return logger_maker



"""Learner."""
from moss.learner.base import BaseLearner
from moss.learner.impala import ImpalaLearner
from moss.learner.ppo import PPOLearner

__all__ = [
  "BaseLearner",
  "ImpalaLearner",
  "PPOLearner",
]



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for aggregating to other loggers."""

import time
from collections import defaultdict
from typing import Callable, Dict, Optional, Sequence

import numpy as np

from moss.utils.loggers import base


class Dispatcher(base.Logger):
  """Writes data to multiple `Logger` objects."""

  def __init__(
    self,
    to: Sequence[base.Logger],
    serialize_fn: Optional[Callable[[base.LoggingData], str]] = None,
  ):
    """Initialize `Dispatcher` connected to several `Logger` objects."""
    self._to = to
    self._serialize_fn = serialize_fn

  def write(self, values: base.LoggingData) -> None:
    """Writes `values` to the underlying `Logger` objects."""
    if self._serialize_fn:
      values = self._serialize_fn(values)  # type: ignore
    for logger in self._to:
      logger.write(values)

  def close(self) -> None:
    """Close."""
    for logger in self._to:
      logger.close()


class TimeAggregator(base.Logger):
  """Logger which writes to another logger at a given time interval."""

  def __init__(self, to: base.Logger, time_delta: float):
    """Initializes the logger.

    Args:
      to: A `Logger` object to which the current object will forward its results
        when `write` is called.
      time_delta: How often to write values out in seconds.
        Note that writes within `time_delta` are stashed.
    """
    self._to = to
    self._values: Dict = defaultdict(list)
    self._time = 0
    self._time_delta = time_delta
    if time_delta < 0:
      raise ValueError(f'time_delta must be greater than 0 (got {time_delta}).')

  def write(self, values: base.LoggingData) -> None:
    """Write."""
    for key, val in values.items():
      self._values[key].append(val)
    now = time.time()
    if (now - self._time) > self._time_delta:
      write_values = {}
      for key in values.keys():
        write_values[key] = np.mean(self._values[key])
        self._values.pop(key)
      self._to.write(write_values)
      self._time = now  # type: ignore

  def close(self) -> None:
    """Close."""
    self._to.close()



"""Queue replay buffer."""
from queue import LifoQueue, Queue

import jax.numpy as jnp
from jax.tree_util import tree_map

from moss.core import Buffer
from moss.types import Trajectory, Transition


class QueueBuffer(Buffer):
  """Queue replay buffer."""

  def __init__(self, maxsize: int, mode: str = "FIFO") -> None:
    """Init.

    Args:
      maxsize: Max size of queue buffer.
      mode: Mode of sample and add data(FIFO or LIFO).
    """
    if mode == "FIFO":
      self._queue: Queue[Trajectory] = Queue(maxsize)
    elif mode == "LIFO":
      self._queue: Queue[Trajectory] = LifoQueue(maxsize)  # type: ignore
    else:
      raise ValueError(f"mode must be `FIFO` or `LIFO`, bug got `{mode}`.")

  def add(self, data: Trajectory) -> None:
    """Add trajectory data to replay buffer.

    Args:
      data: Trajectory data with shape [T, ...].
    """
    self._queue.put(data)

  def sample(self, sample_size: int) -> Transition:
    """Sample trajectory data from replay buffer.

    Returns:
      Batched trajecotry data with shape [T, B, ...].
    """
    data = [self._queue.get() for _ in range(sample_size)]
    stacked_data = tree_map(lambda *x: jnp.stack(x, axis=1), *data)
    return stacked_data



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Logger for values that remain constant."""

from moss.utils.loggers import base


class ConstantLogger(base.Logger):
  """Logger for values that remain constant throughout the experiment.

  This logger is used to log additional values e.g. level_name or
  hyperparameters that do not change in an experiment. Having these values
  allows to group or facet plots when analysing data post-experiment.
  """

  def __init__(
    self,
    constant_data: base.LoggingData,
    to: base.Logger,
  ):
    """Initialise the extra info logger.

    Args:
      constant_data: Key-value pairs containing the constant info to be logged.
      to: The logger to add these extra info to.
    """
    self._constant_data = constant_data
    self._to = to

  def write(self, data: base.LoggingData) -> None:
    """Write."""
    self._to.write({**self._constant_data, **data})

  def close(self) -> None:
    """Close."""
    self._to.close()



"""PettingZoo env wrapper."""
from collections import namedtuple
from typing import Any, Callable, Dict, List, Tuple

import gymnasium
import numpy as np
import tree
from dm_env import StepType, TimeStep
from gymnasium import spaces
from pettingzoo import ParallelEnv
from pettingzoo.utils.wrappers import BaseParallelWrapper

from moss.types import Environment

AgentID = Any
Observation = namedtuple("Observation", ["obs", "info"])


class AutoResetWrapper(BaseParallelWrapper):
  """PettingZoo auto reset wrapper."""

  def __init__(self, env: ParallelEnv) -> None:
    """Init."""
    super().__init__(env)
    self._need_reset: bool = False
    self.agents = self.env.unwrapped.agents

  def step(self, action: Any) -> Tuple[Dict, Dict, Dict, Dict, Dict]:
    """Env step."""
    if self._need_reset:
      self._need_reset = False
      obs, info = self.env.reset()
      reward, terminated, truncated = {}, {}, {}
      for agent_id in self.env.agents:
        reward[agent_id] = 0
        terminated[agent_id] = False
        truncated[agent_id] = False
        assert (
          "first_step" not in info[agent_id]
        ), "info dict cannot contain key 'first_step'."
        info[agent_id]["first_step"] = True
    else:
      obs, reward, terminated, truncated, info = self.env.step(action)
      if all(terminated.values()) or all(truncated.values()):
        self._need_reset = True
      for agent_id in self.env.agents:
        assert (
          "first_step" not in info[agent_id]
        ), "info dict cannot contain key 'first_step'."
        info[agent_id]["first_step"] = False

    return obs, reward, terminated, truncated, info


class ResizeObservation(BaseParallelWrapper):
  """Resize observation wrapper."""

  def __init__(self, env: ParallelEnv, height: int, width: int) -> None:
    """Init."""
    super().__init__(env)
    self.agents = self.env.agents
    self._shape = (height, width)
    obs_type = self.env.unwrapped.obs_type
    if obs_type == "ram":
      observation_space = gymnasium.spaces.Box(
        low=0, high=255, dtype=np.uint8, shape=(128,)
      )
    else:
      num_channels = 3
      if obs_type == "rgb_image":
        num_channels = 3
      elif obs_type == "grayscale_image":
        num_channels = 1
      observation_space = spaces.Box(
        low=0,
        high=255,
        shape=(height, width, num_channels),
        dtype=np.uint8,
      )
    self.env.unwrapped.observation_spaces = {
      agent: observation_space for agent in self.agents
    }

  def reset(self) -> Tuple[Dict, Dict]:
    """Env reset."""
    obs, info = self.env.reset()
    obs = self.resize_observation(obs)
    return obs, info

  def step(self, actions: dict) -> Tuple[Dict, Dict, Dict, Dict, Dict]:
    """Env step."""
    obs, reward, terminated, truncated, info = self.env.step(actions)
    obs = self.resize_observation(obs)
    return obs, reward, terminated, truncated, info

  def resize_observation(self, observation: Dict) -> Dict:
    """Resize observation."""
    try:
      import cv2
    except ImportError as e:
      raise ImportError(
        "opencv (cv2) is not installed, run `pip install opencv-python`"
      ) from e
    observation = tree.map_structure(
      lambda x: cv2.resize(x, self._shape, interpolation=cv2.INTER_AREA),
      observation
    )
    return observation


class PettingZooToDeepmindWrapper(Environment):
  """Wrapper `pettingzoo.ParallelEnv` to `dm_env.Environment`."""

  def __init__(
    self,
    env_fn: Callable,
    height: int = 84,
    width: int = 84,
    **kwargs: Any
  ) -> None:
    """Init."""
    assert (height > 0 and width > 0), (
      f"Expected height and width to be a positive integers, "
      f"got: {height} and {width}"
    )
    # NOTE: AutoReset must be the first wrapper.
    self.env = AutoResetWrapper(env_fn(**kwargs))
    self.env = ResizeObservation(self.env, height, width)
    self._action_spec = {
      agent_id: self.env.action_space(agent_id) for agent_id in self.agents
    }
    self._observation_spec = {
      agent_id: self.env.observation_space(agent_id) for agent_id in self.agents
    }

  def reset(self, **kwargs: Any) -> TimeStep:
    """Env reset."""
    obs, info = self.env.reset(**kwargs)
    for agent_id in self.agents:
      info[agent_id]["agent_id"] = agent_id
    step_type = {agent_id: StepType.FIRST for agent_id in self.env.agents}
    reward = {agent_id: 0. for agent_id in self.env.agents}
    discount = {agent_id: 1. for agent_id in self.env.agents}
    timestep = TimeStep(
      step_type=step_type,
      reward=reward,
      discount=discount,
      observation=Observation(obs, info),
    )
    return timestep

  def step(self, action: Any) -> TimeStep:
    """Env step."""
    obs, reward, terminated, truncated, info = self.env.step(action)
    for agent_id in self.agents:
      info[agent_id]["agent_id"] = agent_id
    step_type, discount = {}, {}
    for agent_id in self.env.agents:
      if terminated[agent_id]:
        step_type[agent_id] = StepType.LAST
        discount[agent_id] = 0.
      elif truncated[agent_id]:
        step_type[agent_id] = StepType.LAST
        discount[agent_id] = 1.
      else:
        step_type[agent_id] =\
          StepType.FIRST if info[agent_id].get("first_step") else StepType.MID
        discount[agent_id] = 1.
    timestep = TimeStep(
      step_type=step_type,
      reward=reward,
      discount=discount,
      observation=Observation(obs, info),
    )
    return timestep

  def action_spec(self) -> Any:
    """Get action spec."""
    return self._action_spec

  def observation_spec(self) -> Any:
    """Get observation spec."""
    return self._observation_spec

  @property
  def agents(self) -> List[str]:
    """Get all agent id."""
    return self.env.agents

  def close(self) -> None:
    """Close env and release resources."""
    self.env.close()



"""Network."""
from moss.network.base import Network
from moss.network.common import CommonNet
from moss.network.ctde import CTDENet
from moss.network.value import DenseValue

__all__ = [
  "CommonNet",
  "CTDENet",
  "DenseValue",
  "Network",
]



"""Atari network."""
from functools import partial
from typing import Any

import numpy as np

from moss.network import CommonNet
from moss.network.action import ActionSpec, DiscreteAction
from moss.network.feature import FeatureSet, FeatureSpec, ImageFeature
from moss.network.feature.encoder import (
  Conv2DConfig,
  ImageFeatureEncoder,
  ResnetConfig,
)
from moss.network.torso import DenseTorso
from moss.network.value import DenseValue

resnet_default_config = [
  ResnetConfig(16, 2),
  ResnetConfig(16, 2),
  ResnetConfig(32, 2),
]
conv2d_default_config = [
  Conv2DConfig(32, 8, 4, "VALID"),
  Conv2DConfig(64, 4, 2, "VALID"),
  Conv2DConfig(64, 3, 1, "VALID"),
]


def network_maker(
  obs_spec: Any,
  action_spec: Any,
  data_format: str = "NHWC",
  use_resnet: bool = False,
  use_orthogonal: bool = True,
) -> Any:
  """Doom network maker."""
  channel, height, width = obs_spec.obs.shape
  num_actions = action_spec.num_values
  doom_frame = FeatureSet(
    name="doom_frame",
    features={
      "frame":
        ImageFeature(
          height, width, channel, data_format, np.int8, "frame",
          lambda x: x / 255.
        )
    },
    encoder_net_maker=lambda: lambda: ImageFeatureEncoder(
      "frame_encoder",
      data_format,
      use_resnet=use_resnet,
      resnet_config=resnet_default_config,
      conv2d_config=conv2d_default_config,
      use_orthogonal=use_orthogonal
    )
  )
  feature_spec = {
    "doom_frame": doom_frame,
  }
  actions = {
    "doom_action":
      DiscreteAction("doom_action", [512], num_actions, use_orthogonal)
  }

  torso_net_maker = partial(DenseTorso, "torso", [512], use_orthogonal)
  value_net_maker = partial(DenseValue, "value", [512, 32], use_orthogonal)
  return CommonNet(
    feature_spec=FeatureSpec(feature_spec),
    action_spec=ActionSpec(actions),
    torso_net_maker=torso_net_maker,
    value_net_maker=value_net_maker,
  )



"""Common feature encoder."""
from typing import Any, List

import haiku as hk
import jax

from moss.types import Array


class CommonEncoder(hk.Module):
  """Common encoder."""

  def __init__(
    self,
    name: str,
    hidden_sizes: List[int],
    use_orthogonal: bool = True
  ) -> None:
    """Init."""
    super().__init__(name)
    self._hidden_sizes = hidden_sizes
    self._use_orthogonal = use_orthogonal

  def __call__(self, inputs: Array) -> Any:
    """Call."""
    layers: List[Any] = []
    w_init = hk.initializers.Orthogonal() if self._use_orthogonal else None
    for hidden_size in self._hidden_sizes:
      layers.append(hk.Linear(hidden_size, w_init=w_init))
      layers.append(jax.nn.relu)
    common_net = hk.Sequential(layers)
    encoder_out = common_net(inputs)
    return encoder_out



"""Base agent."""
from typing import Any, Dict, List, Tuple

import jax
import jax.numpy as jnp
from absl import logging

from moss.core import Agent, Buffer, Predictor
from moss.types import LoggingData, Reward, StepType, TimeStep, Transition
from moss.utils.loggers import Logger


class BaseAgent(Agent):
  """Base agent."""

  def __init__(
    self,
    name: str,
    unroll_length: int,
    buffer: Buffer,
    predictor: Predictor,
    logger: Logger,
  ) -> None:
    """Init."""
    self._name = name
    self._unroll_length = unroll_length
    self._buffer = buffer
    self._predictor = predictor
    self._rnn_state = predictor.initial_state(None)
    self._logger = logger
    self._unroll_steps = 0
    self._trajectory: List[Transition] = []

  def reset(self) -> LoggingData:
    """Agent reset."""
    raise NotImplementedError

  def step(self, timestep: TimeStep) -> Tuple[Dict, Reward]:
    """Agent step."""
    raise NotImplementedError

  def take_action(self, action: Dict[str, Any]) -> Any:
    """Agent take action."""
    raise NotImplementedError

  def inference(self, input_dict: Dict) -> Any:
    """Agent inference."""
    return self._predictor.inference(input_dict, self._rnn_state)

  def result(self, idx: int) -> Any:
    """Get inference result async."""
    action, logits, value, rnn_state = self._predictor.result(idx)
    self._rnn_state = jax.tree_map(lambda x: jnp.array(x), rnn_state)
    return action, logits, value, rnn_state

  def add(self, transition: Transition) -> None:
    """Agent add transition to buffer."""
    is_last_step = transition.step_type == StepType.LAST
    self._trajectory.append(transition)
    self._unroll_steps += 1
    # When unroll_steps == unroll_length + 1 add trajectory to buffer.
    # Because bootstrap need more one step to calculate advantage function.
    if self._unroll_steps > self._unroll_length or is_last_step:
      # Episode end on first trajectory but length less than unroll_length + 1.
      if len(self._trajectory) <= self._unroll_length:
        logging.info(
          "Episode end on first trajectory "
          "but length less than unroll_length + 1."
        )
        self._trajectory = []
        self._unroll_steps = 0
        return

      # Add trajectory to buffer.
      traj = self._trajectory[-(self._unroll_length + 1):]
      stacked_traj = jax.tree_util.tree_map(lambda *x: jnp.stack(x), *traj)
      self._buffer.add(stacked_traj)
      self._trajectory = self._trajectory[-self._unroll_length:]
      self._unroll_steps = 1

      if is_last_step:
        metrics = self.reset()
        self._logger.write(metrics)



"""Value decoder network."""
from typing import Any, List

import haiku as hk
import jax
import jax.numpy as jnp


class DenseValue(hk.Module):
  """Dense value network."""

  def __init__(
    self, name: str, hidden_sizes: List[int], use_orthogonal: bool = True
  ):
    """Init."""
    super().__init__(name)
    self._hidden_sizes = hidden_sizes
    self._use_orthogonal = use_orthogonal

  def __call__(self, inputs: Any) -> Any:
    """Call."""
    w_init = hk.initializers.Orthogonal() if self._use_orthogonal else None
    mlp_layers: List[Any] = []
    for hidden_size in self._hidden_sizes:
      mlp_layers.append(hk.Linear(hidden_size, w_init=w_init))
      mlp_layers.append(jax.nn.relu)
    mlp_layers.append(hk.Linear(1, w_init=w_init))
    value_net = hk.Sequential(mlp_layers)
    value = value_net(inputs)
    value = jnp.squeeze(value, axis=-1)
    return value



"""Action decoder."""
from typing import Any, List, Optional, Type

import distrax
import haiku as hk
import jax
import numpy as np

from moss.network.action.base import Action
from moss.types import Array, KeyArray, SpecArray


class DiscreteAction(Action):
  """Discrete action."""

  def __init__(
    self,
    name: str,
    hidden_sizes: List[int],
    num_actions: int,
    use_orthogonal: bool = True
  ) -> None:
    """Init.

    Args:
      name: Action name.
      hidden_sizes: Hidden sizes of action decoder network.
      num_actions: Discrete action nums.
      use_orthogonal: Whether use orthogonal to initialization params weight.
        Following https://arxiv.org/abs/2006.05990, we set orthogonal
        initialization scale factor of 0.01 for last layer of policy network
        and others layers set as default(1.0).
    """
    self._name = name
    self._hidden_sizes = hidden_sizes
    self._num_actions = num_actions
    self._spec = SpecArray((num_actions,), dtype=np.int8, name=name)
    self._use_orthogonal = use_orthogonal

  def policy_net(self, inputs: Array, mask: Optional[Array] = None) -> Array:
    """Action policy network."""
    w_init = hk.initializers.Orthogonal() if self._use_orthogonal else None
    action_w_init = hk.initializers.Orthogonal(
      scale=0.01
    ) if self._use_orthogonal else None
    layers: List[Any] = []
    for hidden_size in self._hidden_sizes:
      layers.append(hk.Linear(hidden_size, w_init=w_init))
      layers.append(jax.nn.relu)
    layers.append(hk.Linear(self._num_actions, w_init=action_w_init))
    policy_net = hk.Sequential(layers)
    policy_logits = policy_net(inputs)
    if mask is not None:
      policy_logits -= mask * 1e9
    return policy_logits

  @classmethod
  def distribution(
    cls,
    logits: Array,
    temperature: float = 1.,
    dtype: Type = int
  ) -> distrax.Softmax:
    """Action distribution."""
    return distrax.Softmax(logits, temperature, dtype)

  @classmethod
  def sample(
    cls,
    rng: KeyArray,
    logits: Array,
    temperature: float = 1.,
    dtype: Type = int
  ) -> Array:
    """Sample discrete action."""
    distribution = cls.distribution(logits, temperature, dtype)
    return distribution.sample(seed=rng)

  @property
  def name(self) -> Optional[str]:
    """Get action name."""
    return self._name

  @property
  def num_actions(self) -> int:
    """Get discrete action nums."""
    return self._num_actions

  @property
  def spec(self) -> SpecArray:
    """Get action spec."""
    return self._spec



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Timestamp logger."""

import time

from moss.utils.loggers import base


class TimestampLogger(base.Logger):
  """Logger which populates the timestamp key with the current timestamp."""

  def __init__(self, logger: base.Logger, timestamp_key: str):
    """Init."""
    self._logger = logger
    self._timestamp_key = timestamp_key

  def write(self, values: base.LoggingData) -> None:
    """Writer."""
    values = dict(values)
    values[self._timestamp_key] = time.time()
    self._logger.write(values)

  def close(self) -> None:
    """Close."""
    self._logger.close()



"""Feature."""
from typing import Any, Callable, Optional, Tuple, Type

import jax
import jax.numpy as jnp
import numpy as np

from moss.types import SpecArray


class BaseFeature(SpecArray):
  """Base feature."""

  def __init__(
    self,
    shape: Tuple,
    dtype: Type = np.float32,
    name: Optional[str] = None,
    preprocess_fn: Optional[Callable[..., Any]] = None
  ) -> None:
    """Init."""
    super().__init__(shape, dtype, name)
    self._preprocess_fn = preprocess_fn

  def process(self, inputs: Any) -> Any:
    """Feature process."""
    if self._preprocess_fn is not None:
      inputs = self._preprocess_fn(inputs)
    feature = jnp.array(inputs)
    return feature


class ScalarFeature(BaseFeature):
  """Scalar feature."""

  def __init__(
    self,
    dtype: Type = np.float32,
    name: Optional[str] = "ScalarFeature",
    preprocess_fn: Optional[Callable[..., Any]] = None
  ) -> None:
    """Init."""
    super().__init__((), dtype, name, preprocess_fn)

  def process(self, inputs: Any) -> Any:
    """Feature process."""
    if self._process_fn is not None:
      inputs = self._process_fn(inputs)
    feature = jnp.array([inputs])
    return feature


class VectorFeature(BaseFeature):
  """Vector feature."""

  def __init__(
    self,
    length: int,
    dtype: Type = np.float32,
    name: Optional[str] = "VectorFeature",
    preprocess_fn: Optional[Callable[..., Any]] = None
  ) -> None:
    """Init."""
    super().__init__((length,), dtype, name, preprocess_fn)


class OneHotFeature(BaseFeature):
  """One hot feature."""

  def __init__(
    self,
    num_classes: int,
    dtype: Type = np.int8,
    name: Optional[str] = "OneHotFeature",
    preprocess_fn: Optional[Callable[..., Any]] = None
  ) -> None:
    """Init."""
    super().__init__((), dtype, name, preprocess_fn)
    self._num_classes = num_classes

  def process(self, inputs: Any) -> Any:
    """Feature process."""
    if self._preprocess_fn is not None:
      inputs = self._preprocess_fn(inputs)
    feature = jnp.array(inputs)
    return jax.nn.one_hot(feature, num_classes=self.num_classes)

  @property
  def num_classes(self) -> int:
    """Num classes."""
    return self._num_classes


class ImageFeature(BaseFeature):
  """Image feature."""

  def __init__(
    self,
    height: int,
    width: int,
    channel: int,
    data_format: str,
    dtype: Type,
    name: Optional[str] = "ImageFeature",
    preprocess_fn: Optional[Callable[..., Any]] = None
  ) -> None:
    """Init."""
    if data_format == "NHWC":
      shape = (height, width, channel)
    elif data_format == "NCHW":
      shape = (channel, height, width)
    else:
      raise ValueError(
        f"data_format value must be `NHWC` or `NCHW`, but got `{data_format}`."
      )
    super().__init__(shape, dtype, name, preprocess_fn)



"""Base action."""
import abc
from typing import Any, Optional

from distrax import DistributionLike

from moss.types import Array, SpecArray


class Action(abc.ABC):
  """Action."""

  @property
  @abc.abstractmethod
  def name(self) -> SpecArray:
    """Get action name."""

  @property
  @abc.abstractmethod
  def spec(self) -> SpecArray:
    """Action spec."""

  @abc.abstractmethod
  def policy_net(self, inputs: Array, mask: Optional[Array] = None) -> Array:
    """Action policy network."""

  @classmethod
  @abc.abstractmethod
  def distribution(cls, *args: Any, **kwargs: Any) -> DistributionLike:
    """Action distribution."""

  @classmethod
  @abc.abstractmethod
  def sample(cls, *args: Any, **kwargs: Any) -> Any:
    """Sample action."""



# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Transformer module.

Glossary of shapes:
- B: Batch size.
- T: Sequence length.
- D: Model embedding size.
- H: Number of attention heads.
"""
import dataclasses
from typing import Optional

import haiku as hk
import jax
import numpy as np


def _layer_norm(x: jax.Array) -> jax.Array:
  """Applies a unique LayerNorm to `x` with default settings."""
  ln = hk.LayerNorm(axis=-1, create_scale=True, create_offset=True)
  return ln(x)


@dataclasses.dataclass
class Transformer(hk.Module):
  """A transformer stack."""

  num_heads: int  # Number of attention heads.
  num_layers: int  # Number of transformer (attention + MLP) layers to stack.
  attn_size: int  # Size of the attention (key, query, value) vectors.
  dropout_rate: float  # Probability with which to apply dropout.
  widening_factor: int = 4  # Factor by which the MLP hidden layer widens.
  name: Optional[str] = None  # Optional identifier for the module.

  def __call__(
    self,
    embeddings: jax.Array,  # [B, T, D]
    mask: jax.Array,  # [B, T]
  ) -> jax.Array:  # [B, T, D]
    """Transforms input embedding sequences to output embedding sequences."""
    initializer = hk.initializers.VarianceScaling(2 / self.num_layers)
    _, seq_len, model_size = embeddings.shape

    # Compute causal mask for autoregressive sequence modelling.
    mask = mask[:, None, None, :]  # [B, H=1, T'=1, T]
    causal_mask = np.tril(np.ones((1, 1, seq_len, seq_len)))  # [B=1, H=1, T, T]
    mask = mask * causal_mask  # [B, H=1, T, T]

    h = embeddings
    for _ in range(self.num_layers):
      # First the attention block.
      attn_block = hk.MultiHeadAttention(
        num_heads=self.num_heads,
        key_size=self.attn_size,
        model_size=model_size,
        w_init=initializer,
      )
      h_norm = _layer_norm(h)
      h_attn = attn_block(h_norm, h_norm, h_norm, mask=mask)
      h_attn = hk.dropout(hk.next_rng_key(), self.dropout_rate, h_attn)
      h = h + h_attn

      # Then the dense block.
      dense_block = hk.Sequential(
        [
          hk.Linear(self.widening_factor * model_size, w_init=initializer),
          jax.nn.gelu,
          hk.Linear(model_size, w_init=initializer),
        ]
      )
      h_norm = _layer_norm(h)
      h_dense = dense_block(h_norm)
      h_dense = hk.dropout(hk.next_rng_key(), self.dropout_rate, h_dense)
      h = h + h_dense

    return _layer_norm(h)



"""Feature encoder."""
from moss.network.feature.encoder.common import CommonEncoder
from moss.network.feature.encoder.image import (
  Conv2DConfig,
  ImageFeatureEncoder,
  ResnetConfig,
)

__all__ = [
  "CommonEncoder",
  "Conv2DConfig",
  "ResnetConfig",
  "ImageFeatureEncoder",
]



"""Image feature encoder."""
import collections
from typing import Any, List, Optional

import haiku as hk
import jax

from moss.network.layers import ResidualBlock
from moss.types import Array

ResnetConfig = collections.namedtuple(
  "ResnetConfig", ["num_channels", "num_blocks"]
)
Conv2DConfig = collections.namedtuple(
  "Conv2DConfig", ["output_channels", "kernel", "stride", "padding"]
)


class ImageFeatureEncoder(hk.Module):
  """Image featrue encoder."""

  def __init__(
    self,
    name: Optional[str] = None,
    data_format: str = "NHWC",
    use_resnet: bool = False,
    resnet_config: Optional[List[ResnetConfig]] = None,
    conv2d_config: Optional[List[Conv2DConfig]] = None,
    use_orthogonal: bool = True,
  ) -> None:
    """Init.

    Args:
      name: Module name.
      data_format: The data format of the input. Either `NHWC` or `NCHW`. By
        default, `NHWC`.
      use_resnet: Whether use resnet to encoder image feature.
      resnet_config: List of tuple contains 2 nums, num_channels of
        Conv2D and num_blocks of resnet blocks.
      conv2d_config: List of tuple contains 4 arguments (output_channels,
        kernel, stride, padding) of every Conv2D layer.
      use_orthogonal: Whether use orthogonal to initialization params weight.
    """
    super().__init__(name=name)
    self._data_format = data_format
    self._use_resnet = use_resnet
    self._resnet_config = resnet_config
    if self._use_resnet and resnet_config is None:
      raise ValueError(
        "argument `resnet_config` must set when use_resnet is `True`."
      )
    self._conv2d_config = conv2d_config
    if not self._use_resnet and conv2d_config is None:
      raise ValueError(
        "argument `conv2d_config` must set when use_resnet is `False`."
      )
    self._use_orthogonal = use_orthogonal

  def __call__(self, inputs: Array) -> Any:
    """Call."""
    w_init = hk.initializers.Orthogonal() if self._use_orthogonal else None
    if self._use_resnet:
      if self._resnet_config is None:
        raise ValueError(
          "argument `resnet_config` must set when use_resnet is `True`."
        )
      encoder_out = inputs
      for i, (num_channels, num_blocks) in enumerate(self._resnet_config):
        conv = hk.Conv2D(
          num_channels,
          kernel_shape=[3, 3],
          stride=[1, 1],
          w_init=w_init,
          padding="SAME",
          data_format=self._data_format
        )
        encoder_out = conv(encoder_out)  # type: ignore
        encoder_out = hk.max_pool(
          encoder_out,
          window_shape=[1, 3, 3, 1],
          strides=[1, 2, 2, 1],
          padding="SAME"
        )
        for j in range(num_blocks):
          block = ResidualBlock(
            num_channels, "residual_{}_{}".format(i, j), self._data_format,
            self._use_orthogonal
          )
          encoder_out = block(encoder_out)
      encoder_out = hk.Flatten()(encoder_out)
    else:
      if self._conv2d_config is None:
        raise ValueError(
          "argument `conv2d_config` must set when use_resnet is `False`."
        )
      layers: List[Any] = []
      for num_channels, kernel, stride, padding in self._conv2d_config:
        layers.append(
          hk.Conv2D(
            num_channels,
            kernel,
            stride,
            padding=padding,
            w_init=w_init,
            data_format=self._data_format
          )
        )
        layers.append(jax.nn.relu)
      layers.append(hk.Flatten())
      encoder = hk.Sequential(layers)
      encoder_out = encoder(inputs)
    return encoder_out



"""PettingZoo impala example."""
import os
from functools import partial
from typing import Any, Callable

import launchpad as lp
from absl import app, flags, logging
from launchpad.nodes.dereference import Deferred
from launchpad.nodes.python.local_multi_processing import PythonProcess

from examples.pettingzoo.agent import PettingZooAgent
from examples.pettingzoo.network import network_maker
from moss.actor import GenericActor
from moss.buffer import QueueBuffer
from moss.env import DummyVectorEnv
from moss.env.pettingzoo import PettingZooEnv
from moss.learner import ImpalaLearner
from moss.predictor import BasePredictor
from moss.utils.loggers import Logger, experiment_logger_factory
from moss.utils.paths import get_unique_id

flags.DEFINE_string("task_id", "pong_v3", "Task name.")
flags.DEFINE_integer("num_envs", 10, "Num of envs.")
flags.DEFINE_bool(
  "use_resnet", False, "Use resnet or conv2d encoder for petingzoo image."
)
flags.DEFINE_bool(
  "use_orthogonal", True, "Use orthogonal to initialization params weight."
)
flags.DEFINE_integer("num_actors", 16, "Num of actors.")
flags.DEFINE_integer("num_predictors", 2, "Num of predictors.")
flags.DEFINE_integer("unroll_len", 16, "Unroll length.")
flags.DEFINE_integer("predict_batch_size", 32, "Predict batch size.")
flags.DEFINE_integer("training_batch_size", 64, "Training batch size.")
flags.DEFINE_string("model_path", None, "Restore model path.")
flags.DEFINE_integer(
  "publish_interval", 1, "Publish params to predicotrs interval(by train steps)."
)
flags.DEFINE_integer("save_interval", 500, "Save interval(by train steps).")
flags.DEFINE_float("learning_rate", 5e-4, "Learning rate.")
flags.DEFINE_float("gamma", 0.99, "Reward discount rate.")
flags.DEFINE_float("gae_lambda", 0.95, "GAE lambda.")
flags.DEFINE_float("rho_clip", 0.9, "Clip threshold for importance ratios.")
flags.DEFINE_float(
  "pg_rho_clip", 0.9, "Clip threshold for policy gradient importance ratios."
)
flags.DEFINE_float("critic_coef", 0.25, "Critic coefficient.")
flags.DEFINE_float("entropy_coef", 0.01, "Entropy coefficient.")
flags.DEFINE_integer("buffer_size", 2048, "Replay buffer size.")
flags.DEFINE_string(
  "buffer_mode", "FIFO", "Queue buffer mode(Support `FIFO` and `LIFO`)."
)
flags.DEFINE_string("lp_launch_type", "local_mp", "Launch type.")

FLAGS = flags.FLAGS


def make_lp_program() -> Any:
  """Make launchpad program."""
  (exp_uid,) = get_unique_id()
  task_id = FLAGS.task_id
  num_envs = FLAGS.num_envs
  unroll_len = FLAGS.unroll_len

  dummy_env = PettingZooEnv(task_id)
  obs_spec: Any = dummy_env.observation_spec()
  action_spec: Any = dummy_env.action_spec()
  use_resnet = FLAGS.use_resnet
  use_orthogonal = FLAGS.use_orthogonal

  logging.info(f"Task id: {task_id}")
  logging.info(f"Observation spec: {obs_spec}")
  logging.info(f"Action spec: {action_spec}")

  def env_maker() -> Any:
    """Env maker."""
    return DummyVectorEnv(num_envs, PettingZooEnv, task_id=task_id)

  def agent_maker(buffer: QueueBuffer, predictor: BasePredictor) -> Callable:
    """Agent maker."""

    def agent_wrapper(timestep: Any, logger: Logger) -> PettingZooAgent:
      """Return a agent."""
      del timestep
      return PettingZooAgent(unroll_len, buffer, predictor, logger)

    return agent_wrapper

  logger_fn = experiment_logger_factory(
    project=task_id, uid=exp_uid, time_delta=2.0, print_fn=logging.info
  )

  program = lp.Program("impala")
  with program.group("buffer"):
    buffer_node = lp.CourierNode(
      QueueBuffer,
      FLAGS.buffer_size,
      FLAGS.buffer_mode,
    )
    buffer = program.add_node(buffer_node)

  predictors = []
  with program.group("predictor"):
    for _ in range(FLAGS.num_predictors):
      predictor_node = lp.CourierNode(
        BasePredictor,
        FLAGS.predict_batch_size,
        partial(
          network_maker, obs_spec, action_spec, "NHWC", use_resnet,
          use_orthogonal
        ),
        logger_fn,
      )
      predictor = program.add_node(predictor_node)
      predictors.append(predictor)

  with program.group("actor"):
    for i in range(FLAGS.num_actors):
      actor_node = lp.CourierNode(
        GenericActor,
        Deferred(agent_maker, buffer, predictors[i % FLAGS.num_predictors]),
        env_maker,
        logger_fn,
      )
      program.add_node(actor_node)

  with program.group("learner"):
    save_path = os.path.join("checkpoints", task_id, exp_uid)
    learner_node = lp.CourierNode(
      ImpalaLearner,
      buffer,
      predictors,
      partial(
        network_maker, obs_spec, action_spec, "NHWC", use_resnet, use_orthogonal
      ),
      logger_fn,
      batch_size=FLAGS.training_batch_size,
      save_interval=FLAGS.save_interval,
      save_path=save_path,
      model_path=FLAGS.model_path,
      publish_interval=FLAGS.publish_interval,
      learning_rate=FLAGS.learning_rate,
      discount=FLAGS.gamma,
      gae_lambda=FLAGS.gae_lambda,
      clip_rho_threshold=FLAGS.rho_clip,
      clip_pg_rho_threshold=FLAGS.pg_rho_clip,
      critic_coef=FLAGS.critic_coef,
      entropy_coef=FLAGS.entropy_coef,
    )
    program.add_node(learner_node)
  return program


def main(_):
  """Main function."""
  program = make_lp_program()
  local_resources = {
    "actor":
      PythonProcess(
        env={
          "CUDA_VISIBLE_DEVICES": "",
          "XLA_PYTHON_CLIENT_PREALLOCATE": "false"
        }
      ),
    "learner":
      PythonProcess(
        env={
          "CUDA_VISIBLE_DEVICES": "0",
          "XLA_PYTHON_CLIENT_PREALLOCATE": "false"
        }
      ),
    "predictor":
      PythonProcess(
        env={
          "CUDA_VISIBLE_DEVICES": "0",
          "XLA_PYTHON_CLIENT_PREALLOCATE": "false"
        }
      ),
    "buffer":
      PythonProcess(
        env={
          "CUDA_VISIBLE_DEVICES": "",
          "XLA_PYTHON_CLIENT_PREALLOCATE": "false"
        }
      ),
  }
  lp.launch(program, local_resources=local_resources, terminal="tmux_session")


if __name__ == "__main__":
  os.environ["CUDA_VISIBLE_DEVICES"] = ""
  app.run(main)



"""Types."""
import collections
from typing import Any, Dict, List, NamedTuple, Union

import haiku as hk
import jax
import numpy as np
import optax
from dm_env import Environment, StepType, TimeStep  # noqa: F401
from dm_env.specs import Array as SpecArray  # noqa: F401
from jax.random import KeyArray  # noqa: F401

Array = Union[np.ndarray, jax.Array]
Params = Union[hk.Params, optax.Params]
OptState = optax.OptState
LoggingData = Dict[str, Any]

Observation = Any
AgentState = Any
RNNState = Any
Action = Any
Reward = Any
History = Any


class Transition(NamedTuple):
  """Transtion."""
  step_type: StepType
  input_dict: Dict
  action: Action
  rnn_state: RNNState
  reward: Reward
  policy_logits: Dict[str, Array]
  behaviour_value: Array


Trajectory = Union[Transition, List[Transition]]
NetOutput = collections.namedtuple(
  "NetOutput", ["policy_logits", "value", "rnn_state"]
)



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities to use within loggers."""

import queue
import threading
from typing import Callable, Generic, TypeVar

from absl import logging

E = TypeVar("E")


class AsyncExecutor(Generic[E]):
  """Executes a blocking function asynchronously on a queue of items."""

  def __init__(
    self,
    fn: Callable[[E], None],
    queue_size: int = 1,
    interruptible_interval_secs: float = 1.0,
  ):
    """Buffers elements in a queue and runs `fn` asynchronously..

    NOTE: Once closed, `AsyncExecutor` will block until current `fn` finishes
      but is not guaranteed to dequeue all elements currently stored in
      the data queue. This is intentional so as to prevent a blocking `fn` call
      from preventing `AsyncExecutor` from closing.

    Args:
      fn: A callable to be executed upon dequeuing an element from data
        queue.
      queue_size: The maximum size of the synchronized buffer queue.
      interruptible_interval_secs: Timeout interval in seconds for blocking
        queue operations after which the background threads check for errors and
        if background threads should stop.
    """
    self._data = queue.Queue(maxsize=queue_size)  # type: ignore
    self._should_stop = threading.Event()
    self._errors = queue.Queue()  # type: ignore
    self._interruptible_interval_secs = interruptible_interval_secs

    def _dequeue() -> None:
      """Dequeue data from a queue and invoke blocking call."""
      while not self._should_stop.is_set():
        try:
          element = self._data.get(timeout=self._interruptible_interval_secs)
          # Execute fn upon dequeuing an element from the data queue.
          fn(element)
        except queue.Empty:
          # If queue is Empty for longer than the specified time interval,
          # check again if should_stop has been requested and retry.
          continue
        except Exception as e:
          logging.error("AsyncExecuter thread terminated with error.")
          logging.exception(e)
          self._errors.put(e)
          self._should_stop.set()
          raise  # Never caught by anything, just terminates the thread.

    self._thread = threading.Thread(target=_dequeue, daemon=True)
    self._thread.start()

  def _raise_on_error(self) -> None:
    try:
      # Raise the error on the caller thread if an error has been raised in the
      # looper thread.
      raise self._errors.get_nowait()
    except queue.Empty:
      pass

  def close(self) -> None:
    """Close."""
    self._should_stop.set()
    # Join all background threads.
    self._thread.join()
    # Raise errors produced by background threads.
    self._raise_on_error()

  def put(self, element: E) -> None:
    """Puts `element` asynchronuously onto the underlying data queue.

    The write call blocks if the underlying data_queue contains `queue_size`
      elements for over `self._interruptible_interval_secs` second, in which
      case we check if stop has been requested or if there has been an error
      raised on the looper thread. If neither happened, retry enqueue.

    Args:
      element: an element to be put into the underlying data queue and dequeued
        asynchronuously for `fn(element)` call.
    """
    while not self._should_stop.is_set():
      try:
        self._data.put(element, timeout=self._interruptible_interval_secs)
        break
      except queue.Full:
        continue
    else:
      # If `should_stop` has been set, then raises if any has been raised on
      # the background thread.
      self._raise_on_error()



"""LSTM torso network."""
from typing import Any, Dict, Optional, Tuple

import haiku as hk
import jax.numpy as jnp

from moss.types import Array


class LSTMTorso(hk.RNNCore):
  """LSTM torso network."""

  def __init__(self, name: str, hidden_size: int):
    """Init."""
    super().__init__(name)
    self._core = hk.LSTM(hidden_size)

  def __call__(self, inputs: Dict[str, Any],
               state: hk.LSTMState) -> Tuple[Array, hk.LSTMState]:
    """Call."""
    torso_input = jnp.concatenate(list(inputs.values()), axis=-1)
    torso_out, new_state = self._core(torso_input, state)
    return torso_out, new_state

  def initial_state(self, batch_size: Optional[int]) -> hk.LSTMState:
    """Constructs an initial state for this core."""
    return self._core.initial_state(batch_size)



"""Base predictor."""
import queue
import threading
import time
from concurrent.futures import Future
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Tuple

import jax
import jax.numpy as jnp
from absl import logging

from moss.core import Params, Predictor
from moss.network import Network
from moss.types import Array, KeyArray, NetOutput, RNNState
from moss.utils.loggers import Logger


class BasePredictor(Predictor):
  """Base predictor."""

  def __init__(
    self,
    batch_size: int,
    network_maker: Callable[[], Network],
    logger_fn: Callable[..., Logger],
    seed: int = 42,
  ) -> None:
    """Init.

    Args:
      batch_size: Predict batch size.
      network_maker: Network maker function.
      logger_fn: Logger function.
      seed: random seed.
    """
    self._batch_size = batch_size
    self._network = network_maker()
    self._logger = logger_fn(label="Predictor")
    self._rng = jax.random.PRNGKey(seed)
    self._requests: queue.Queue[Tuple[Dict, RNNState, Future]] = queue.Queue()
    self._results: Dict[int, Future] = {}
    self._resp_id: int = 0
    self._params: Optional[Params] = None
    self._params_mutex = threading.Lock()
    self._params_initialized = threading.Condition(self._params_mutex)
    self._inference_mutex = threading.Lock()
    logging.info(jax.devices())

  @partial(jax.jit, static_argnums=0)
  def _forward(
    self, params: Params, input_dict: Dict, rnn_state: RNNState, rng: KeyArray
  ) -> Tuple[Dict[str, Array], NetOutput]:
    """Forward."""
    action, net_output = self._network.forward(
      params, input_dict, rnn_state, rng, False
    )
    return action, net_output

  def _batch_request(self) -> Tuple[Dict, RNNState, List[Future]]:
    """Get batch request data."""
    input_dicts: List[Any] = []
    rnn_states: List[Any] = []
    futures: List[Any] = []
    while len(input_dicts) < self._batch_size:
      try:
        # The function of timeout is to ensure that there
        # is at least one vaild data in input_dicts.
        timetout = 0.05 if len(input_dicts) > 0 else None
        request = self._requests.get(timeout=timetout)
        input_dict, rnn_state, future = request
        input_dicts.append(input_dict)
        rnn_states.append(rnn_state)
        futures.append(future)
      except queue.Empty:
        logging.info("Get batch request timeout.")
        padding_len = self._batch_size - len(input_dicts)
        state_padding = jax.tree_util.tree_map(
          lambda x: jnp.zeros_like(x), input_dicts[0]
        )
        rnn_state_padding = jax.tree_util.tree_map(
          lambda x: jnp.zeros_like(x), rnn_states[0]
        )
        for _ in range(padding_len):
          input_dicts.append(state_padding)
          rnn_states.append(rnn_state_padding)
        break
    batch_input_dict = jax.tree_util.tree_map(
      lambda *x: jnp.stack(x), *input_dicts
    )
    batch_rnn_state = jax.tree_util.tree_map(
      lambda *x: jnp.stack(x), *rnn_states
    )
    return batch_input_dict, batch_rnn_state, futures

  def initial_state(self, batch_size: Optional[int]) -> Any:
    """Get initial state."""
    return self._network.initial_state(batch_size)

  def update_params(self, params: Params) -> None:
    """Update params by Learner."""
    with self._params_mutex:
      if self._params is None:
        self._params = params
        self._params_initialized.notify_all()
      else:
        self._params = params

  def inference(self, input_dict: Dict, rnn_state: RNNState) -> int:
    """Inference.

    Args:
      input_dict: processed by Agent, and input to Network.
      rnn_state: the last rnn state input to Network.
    """
    with self._inference_mutex:
      self._resp_id += 1
      resp_id = self._resp_id
    future: Future = Future()
    self._results[resp_id] = future
    self._requests.put((input_dict, rnn_state, future))
    return resp_id

  def result(self, id: int) -> Any:
    """Get result async."""
    future = self._results.pop(id)
    result = future.result()
    return result

  def run(self) -> None:
    """Run predictor."""
    with self._params_initialized:
      if self._params is None:
        self._params_initialized.wait()
    rng = self._rng
    while True:
      get_batch_req_start = time.time()
      batch_input_dict, batch_rnn_state, futures = self._batch_request()
      get_batch_req_time = time.time() - get_batch_req_start

      forward_start = time.time()
      rng, sub_rng = jax.random.split(rng)
      action, net_output = self._forward(
        self._params, batch_input_dict, batch_rnn_state, sub_rng
      )
      (action, net_output) = jax.device_get((action, net_output))
      forward_time = time.time() - forward_start

      for i, future in enumerate(futures):
        action_i = jax.tree_map(lambda x: x[i], action)  # noqa: B023
        net_output_i = jax.tree_map(lambda x: x[i], net_output)  # noqa: B023
        result = (
          action_i, net_output_i.policy_logits, net_output_i.value,
          net_output_i.rnn_state
        )
        future.set_result(result)

      metrics = {
        "time/get batch": get_batch_req_time,
        "time/batch forward": forward_time,
      }
      self._logger.write(metrics)



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for logging to a TensorboardX SummaryWriter."""

import datetime
import time
from typing import Optional

from absl import logging
from tensorboardX import SummaryWriter

from moss.utils.loggers import base


def _format_key(key: str) -> str:
  """Internal function for formatting keys in TensorboardX format."""
  return key.title().replace('_', '').replace(' ', '')


class TBXSummaryLogger(base.Logger):
  """Logs to a TensorboardX SummaryWriter created in a given logdir.

  If multiple TBXSummaryLogger are created with the same logdir, results will be
  categorized by labels.
  """

  def __init__(
    self, logdir: str, label: str = 'Logs', steps_key: Optional[str] = None
  ):
    """Initializes the logger.

    Args:
      logdir: directory to which we should log files.
      label: label string to use when logging. Default to 'Logs'.
      steps_key: key to use for steps. Must be in the values passed to write.
    """
    self._time = time.time()
    self.label = label
    self._iter = 0
    filename_suffix = f".{datetime.datetime.now().strftime('%f')}"
    self.summary = SummaryWriter(logdir, filename_suffix=filename_suffix)
    self._steps_key = steps_key

  def write(self, values: base.LoggingData) -> None:
    """Write."""
    if self._steps_key is not None and self._steps_key not in values:
      logging.warning('steps key %s not found. Skip logging.', self._steps_key)
      return

    step = values[self._steps_key] if self._steps_key is not None else self._iter

    # TODO(b/159065169): Remove this suppression once the bug is resolved.
    # pytype: disable=unsupported-operands
    for key in values.keys() - [self._steps_key]:
      # pytype: enable=unsupported-operands
      self.summary.add_scalar(
        f'{self.label}/{_format_key(key)}', values[key], step
      )
    self._iter += 1

  def close(self) -> None:
    """Close summary."""
    self.summary.close()



"""Vector environment."""
from moss.env.vector.base import (
  BaseEnvWorker,
  DummyVectorEnv,
  EnvpoolVectorEnv,
  RayVectorEnv,
)

__all__ = [
  "BaseEnvWorker",
  "DummyVectorEnv",
  "EnvpoolVectorEnv",
  "RayVectorEnv",
]



"""Envpool environment suite moss."""
from typing import Any, Dict, List

import envpool
import jax
import numpy as np

from moss.env.base import BaseEnv
from moss.types import Array, TimeStep

AgentID = Any


class EnvpoolEnv(BaseEnv):
  """Envpool env."""

  def __init__(self, task_id: str, **kwargs: Any) -> None:
    """Init."""
    self._env = envpool.make_dm(task_id, **kwargs)

  def reset(self) -> Dict[AgentID, TimeStep]:
    """Reset."""
    timestep = self._env.reset()
    return self.timestep_process(timestep)

  def step(self, action: Any) -> Dict[AgentID, TimeStep]:
    """Step."""
    action = self.action_process(action)
    timestep = self._env.step(action)
    return self.timestep_process(timestep)

  def timestep_process(self, timestep: TimeStep) -> Dict[AgentID, TimeStep]:
    """Timestep process."""

    def split_batch_timestep(batch: TimeStep) -> List[TimeStep]:
      """Split batch timestep by env id."""
      size = batch.step_type.size
      timesteps = [
        jax.tree_util.tree_map(lambda x: x[i], batch)  # noqa: B023
        for i in range(size)
      ]
      return timesteps

    timesteps = split_batch_timestep(timestep)
    new_timesteps = {
      agent_id: timestep for agent_id, timestep in enumerate(timesteps)
    }
    return new_timesteps

  def action_process(self, action: Dict[AgentID, Any]) -> Array:
    """Action process."""
    return np.stack(list(action.values()))

  def action_spec(self) -> Any:
    """Get action spec."""
    return self._env.action_spec()

  def observation_spec(self) -> Any:
    """Get observation spec."""
    return self._env.observation_spec()

  def close(self) -> None:
    """Close env and release resources."""
    self._env.close()



"""Vectorized environments."""
from functools import partial
from typing import Any, Callable, Dict

from moss.env.base import BaseEnv
from moss.env.envpool import EnvpoolEnv
from moss.env.vector.worker import BaseEnvWorker, DummyWorker, RayEnvWorker
from moss.types import TimeStep

EnvID = Any
AgentID = Any


class BaseVectorEnv(object):
  """Base vectorized environments.

  Provide a universal interface to support both vectorized and multi-agent
  environment by use a generic env return format in
  `Dict[EnvID, Dict[AgentID, TimeStep]]`.
  """

  def __init__(
    self,
    num_envs: int,
    env_maker: Callable[[], BaseEnv],
    worker_fn: Callable[[Callable[[], BaseEnv]], BaseEnvWorker],
    **kwargs: Any,
  ) -> None:
    """Init.

    Args:
      num_envs: Num of vectorized environments.
      env_maker: Environment maker function.
      worker_fn: Environment worker function.
      kwargs: Any others env arguments.
    """
    self._num_envs = num_envs
    self._env_maker = env_maker
    self._workers = [worker_fn(env_maker, **kwargs) for _ in range(num_envs)]

  def reset(self) -> Dict[EnvID, Dict[AgentID, TimeStep]]:
    """Resets all environments and returns the initial timestep.

    This method is particularly useful when working with vectorized
    environments,where multiple instances of an environment are being managed
    simultaneously.It ensures that each environment is started from the
    beginning, typically after the end of an episode or during the start of a
    new experiment.

    Returns:
      A dictionary mapping each environment ID (EnvID) to another dictionary,
      which in turn maps each agent ID (AgentID) within that environment to the
      initial time step (TimeStep).The TimeStep object typically includes the
      initial observation, the initial reward (often zero in the first step),
      and other relevant time step information such as whether the state is
      terminal (usually False at the start).

    Example:
      >>> envs = BaseVectorEnv()
      >>> timestpes_dict = envs.reset()
      >>> for env_id, timesteps in timestpes_dict.items():
      ...   for agent_id, timestep in timesteps.items():
      ...     # process the initial timestep for each agent in each environment
      ...     pass
    """
    timesteps_dict = {
      env_id: worker.reset() for env_id, worker in enumerate(self._workers)
    }
    return timesteps_dict

  def step(
    self, action_dict: Dict[EnvID, Dict[AgentID, Any]]
  ) -> Dict[EnvID, Dict[AgentID, TimeStep]]:
    """Vectorized environments step.

    Args:
      action_dict: A dictionary mapping each environment ID (EnvID) to action
      which is a NumPy array, or a nested dict, list or tuple of arrays
      corresponding to `action_spec()`.

    Returns:
      A dictionary mapping each environment ID (EnvID) to another dictionary,
      which in turn maps each agent ID (AgentID) within that environment to the
      next timestep (TimeStep).The TimeStep object typically includes the
      next observation, the next reward (often zero in the first step),
      and other relevant time step information such as whether the state is
      terminal (usually False at the start).
    """
    timesteps_dict = {
      env_id: worker.step(action_dict[env_id])
      for env_id, worker in enumerate(self._workers)
    }
    return timesteps_dict

  def observation_spec(self) -> Any:
    """Defines the observations provided by the environment.

    May use a subclass of `specs.Array` that specifies additional properties
    such as min and max bounds on the values.

    Returns:
      An `Array` spec, or a nested dict, list or tuple of `Array` specs.
    """
    pass

  def action_spec(self) -> Any:
    """Defines the actions that should be provided to `step`.

    May use a subclass of `specs.Array` that specifies additional properties
    such as min and max bounds on the values.

    Returns:
      An `Array` spec, or a nested dict, list or tuple of `Array` specs.
    """
    pass

  @property
  def num_envs(self) -> int:
    """Num of vectorized environments."""
    return self._num_envs


class DummyVectorEnv(BaseVectorEnv):
  """Dummy vectorized environments wrapper, implemented in for-loop."""

  def __init__(
    self, num_envs: int, env_maker: Callable[[], BaseEnv], **kwargs: Any
  ) -> None:
    """Dummy vectorized environments wrapper."""
    super().__init__(num_envs, env_maker, DummyWorker, **kwargs)


class EnvpoolVectorEnv(BaseVectorEnv):
  """Envpool vectorized environments warrper, implemented via `envpool`."""

  def __init__(self, task_id: str, **kwargs: Any) -> None:
    """Envpool vectorized environments warrper."""
    env_maker = partial(EnvpoolEnv, task_id, **kwargs)
    super().__init__(1, env_maker, DummyWorker)

  @property
  def num_envs(self) -> int:
    """Num of vectorized environments."""
    return self._workers[0].env.config["num_envs"]


class RayVectorEnv(BaseVectorEnv):
  """Ray vectorized environments wrapper, implemented in ray."""

  def __init__(
    self, num_envs: int, env_maker: Callable[[], BaseEnv], **kwargs: Any
  ) -> None:
    """Ray vectorized environments wrapper."""
    super().__init__(num_envs, env_maker, RayEnvWorker, **kwargs)



"""Custom environments."""
from moss.env.base import BaseEnv
from moss.env.envpool import EnvpoolEnv
from moss.env.vector.base import (
  BaseVectorEnv,
  DummyVectorEnv,
  EnvpoolVectorEnv,
  RayVectorEnv,
)
from moss.env.vector.worker import BaseEnvWorker, DummyWorker, RayEnvWorker

__all__ = [
  "BaseEnv",
  "BaseVectorEnv",
  "DummyVectorEnv",
  "RayVectorEnv",
  "EnvpoolEnv",
  "EnvpoolVectorEnv",
  "BaseEnvWorker",
  "DummyWorker",
  "RayEnvWorker",
]
# Internal imports.



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Moss loggers."""

from moss.utils.loggers.aggregators import Dispatcher, TimeAggregator
from moss.utils.loggers.asynchronous import AsyncLogger
from moss.utils.loggers.base import (
  Logger,
  LoggerFactory,
  LoggerLabel,
  LoggerStepsKey,
  LoggingData,
  NoOpLogger,
  TaskInstance,
  to_numpy,
)
from moss.utils.loggers.constant import ConstantLogger
from moss.utils.loggers.csv import CSVLogger
from moss.utils.loggers.dataframe import InMemoryLogger
from moss.utils.loggers.default import (
  experiment_logger_factory,  # pylint: disable=g-bad-import-order
)
from moss.utils.loggers.default import (
  make_default_logger,  # pylint: disable=g-bad-import-order
)
from moss.utils.loggers.filters import (
  GatedFilter,
  KeyFilter,
  NoneFilter,
  TimeFilter,
)
from moss.utils.loggers.flatten import FlattenDictLogger
from moss.utils.loggers.terminal import TerminalLogger
from moss.utils.loggers.timestamp import TimestampLogger

__all__ = [
  "TimeAggregator",
  "Dispatcher",
  "AsyncLogger",
  "Logger",
  "LoggerFactory",
  "LoggerLabel",
  "LoggerStepsKey",
  "LoggingData",
  "NoOpLogger",
  "TaskInstance",
  "to_numpy",
  "ConstantLogger",
  "CSVLogger",
  "InMemoryLogger",
  "experiment_logger_factory",
  "make_default_logger",
  "GatedFilter",
  "KeyFilter",
  "NoneFilter",
  "TimeFilter",
  "FlattenDictLogger",
  "TerminalLogger",
  "TimestampLogger",
]
# Internal imports.



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base logger."""

import abc
from typing import Any, Mapping, Optional

import numpy as np
import tree
from typing_extensions import Protocol

LoggingData = Mapping[str, Any]


class Logger(abc.ABC):
  """A logger has a `write` method."""

  @abc.abstractmethod
  def write(self, data: LoggingData) -> None:
    """Writes `data` to destination (file, terminal, database, etc)."""

  @abc.abstractmethod
  def close(self) -> None:
    """Closes the logger, not expecting any further write."""


TaskInstance = int
# TODO(stanczyk): Turn LoggerLabel into an enum of [Learner, Actor, Evaluator].
LoggerLabel = str
LoggerStepsKey = str


class LoggerFactory(Protocol):
  """Logger factory."""

  def __call__(
    self,
    label: LoggerLabel,
    steps_key: Optional[LoggerStepsKey] = None,
    instance: Optional[TaskInstance] = None
  ) -> Logger:
    """Call."""
    ...


class NoOpLogger(Logger):
  """Simple Logger which does nothing and outputs no logs.

  This should be used sparingly, but it can prove useful if we want to quiet an
  individual component and have it produce no logging whatsoever.
  """

  def write(self, data: LoggingData) -> Any:
    """Writer."""
    pass

  def close(self) -> Any:
    """Close."""
    pass


def tensor_to_numpy(value: Any) -> Any:
  """Tensor to numpy."""
  if hasattr(value, 'numpy'):
    return value.numpy()  # tf.Tensor (TF2).
  if hasattr(value, 'device_buffer'):
    return np.asarray(value)  # jnp.DeviceArray.
  return value


def to_numpy(values: Any) -> Any:
  """Converts tensors in a nested structure to numpy.

  Converts tensors from TensorFlow to Numpy if needed without importing TF
  dependency.

  Args:
    values: nested structure with numpy and / or TF tensors.

  Returns:
    Same nested structure as values, but with numpy tensors.
  """
  return tree.map_structure(tensor_to_numpy, values)



"""Base network."""
import abc
from typing import Any, Dict, Optional, Tuple

from moss.network.action import ActionSpec
from moss.types import Array, KeyArray, NetOutput, Params, RNNState


class Network(abc.ABC):
  """Neural network interface."""

  @property
  @abc.abstractmethod
  def action_spec(self) -> ActionSpec:
    """Action spec."""

  @abc.abstractmethod
  def initial_params(self, rng: KeyArray) -> Params:
    """Init network's params."""

  @abc.abstractmethod
  def initial_state(self, batch_size: Optional[int]) -> Any:
    """Constructs an initial state for rnn core."""

  @abc.abstractmethod
  def forward(
    self, params: Params, input_dict: Dict, rnn_state: RNNState, rng: KeyArray,
    training: bool
  ) -> Tuple[Dict[str, Array], NetOutput]:
    """Network forward."""



"""Feature spec."""
from typing import Any, Dict

from moss.network.feature.feature_set import FeatureSet
from moss.types import SpecArray


class FeatureSpec(object):
  """Feature spec."""

  def __init__(self, feature_sets: Dict[str, FeatureSet]) -> None:
    """Init."""
    self._feature_sets = feature_sets

  @property
  def feature_sets(self) -> Dict[str, FeatureSet]:
    """Feature sets."""
    return self._feature_sets

  @property
  def spec(self) -> Dict[str, Dict[str, SpecArray]]:
    """Feature spec."""
    feature_spec = {
      name: feature_set.spec for name, feature_set in self._feature_sets.items()
    }
    return feature_spec

  def generate_value(self) -> Dict[str, Any]:
    """Generate a test value which conforms to this feature spec."""
    value = {
      name: feature_set.generate_value()
      for name, feature_set in self._feature_sets.items()
    }
    return value



"""Core interface."""
import abc
from typing import Any, Dict, Optional, Tuple

from moss.env.base import TimeStep
from moss.types import (
  Array,
  LoggingData,
  OptState,
  Params,
  Reward,
  RNNState,
  Trajectory,
  Transition,
)


class Worker(abc.ABC):
  """Worker interface."""

  @abc.abstractmethod
  def run(self) -> None:
    """Runs the worker."""


class Actor(Worker):
  """Actor interface."""


class Agent(abc.ABC):
  """Agent interface."""

  @abc.abstractmethod
  def reset(self) -> Any:
    """Reset agent."""

  @abc.abstractmethod
  def step(self, timestep: TimeStep) -> Tuple[Dict, Reward]:
    """Take step."""

  @abc.abstractmethod
  def inference(self, input_dict: Dict) -> Any:
    """Inference."""

  @abc.abstractmethod
  def result(self, idx: int) -> Any:
    """Get inference result async."""

  @abc.abstractmethod
  def take_action(self, action: Dict[str, Any]) -> Any:
    """Take action."""

  @abc.abstractmethod
  def add(self, transition: Transition) -> None:
    """Add transition to buffer."""


class Buffer(abc.ABC):
  """Replay buffer interface."""

  @abc.abstractmethod
  def add(self, traj: Trajectory) -> None:
    """Add replay transtion."""

  @abc.abstractmethod
  def sample(self, sample_size: int) -> Transition:
    """Sample data."""


class Learner(Worker):
  """RL learner interface."""

  @abc.abstractmethod
  def _loss(self, params: Params, data: Transition) -> Tuple[Array, LoggingData]:
    """Loss function."""

  @abc.abstractmethod
  def _train_step(
    self, params: Params, opt_state: OptState, data: Transition
  ) -> Any:
    """Training step."""

  @abc.abstractmethod
  def _publish_params(self, params: Params) -> Any:
    """Publish params to pedictors."""

  @abc.abstractmethod
  def _load_model(self, model_path: str) -> Any:
    """Load model."""

  @abc.abstractmethod
  def _save_model(self, save_path: str, params: Params) -> None:
    """Save model."""


class Predictor(Worker):
  """Predictor interface."""

  @abc.abstractmethod
  def update_params(self, params: Params) -> None:
    """Update params."""

  @abc.abstractmethod
  def inference(self, input_dict: Dict, rnn_state: RNNState) -> Any:
    """Inference."""

  @abc.abstractmethod
  def result(self, idx: int) -> Any:
    """Get result async."""

  @abc.abstractmethod
  def initial_state(self, batch_size: Optional[int]) -> RNNState:
    """Get initial rnn state."""



"""Generic actor."""
import collections
import time
from typing import Callable, Dict, Tuple

import jax
from absl import logging

from moss.agent import BaseAgent
from moss.core import Actor
from moss.env import BaseVectorEnv
from moss.types import Transition
from moss.utils.loggers import Logger


class GenericActor(Actor):
  """A generic actor implemention.

  A generic actor for all types of environments:
    - Single agent environment.
    - Multi agent environment.
    - Single environment.
    - Vectorized environment.
  """

  def __init__(
    self,
    agent_maker: Callable[..., BaseAgent],
    env_maker: Callable[[], BaseVectorEnv],
    logger_fn: Callable[..., Logger],
  ) -> None:
    """Init."""
    self._agent_maker = agent_maker
    self._env_maker = env_maker
    self._logger_fn = logger_fn
    self._logger = logger_fn(label="Actor")
    logging.info(jax.devices())

  def run(self) -> None:
    """Run actor."""
    agent_logger = self._logger_fn(label="Agent")
    agents: Dict[Tuple[int, int], BaseAgent] = {}
    envs = self._env_maker()
    timesteps_dict = envs.reset()
    while True:
      actor_step_start = time.time()
      input_dicts = collections.defaultdict(list)
      rewards = collections.defaultdict(list)
      responses = collections.defaultdict(list)
      for env_id, timesteps in timesteps_dict.items():
        for player_id, timestep in timesteps.items():
          # NOTE: To avoid confusion between environment agent and rl agent,
          # we refer to the agent_id as a player_id.
          ep_id = (env_id, player_id)
          if ep_id not in agents.keys():
            agents[ep_id] = self._agent_maker(timestep, agent_logger)
          input_dict, reward = agents[ep_id].step(timestep)
          response = agents[ep_id].inference(input_dict)
          input_dicts[env_id].append(input_dict)
          rewards[env_id].append(reward)
          responses[env_id].append(response)
      get_result_time = 0.0
      action_dict = collections.defaultdict(dict)  # type: ignore
      for env_id, timesteps in timesteps_dict.items():
        for (player_id, timestep), input_dict, response, reward in zip(
          timesteps.items(), input_dicts[env_id], responses[env_id],
          rewards[env_id]
        ):
          ep_id = (env_id, player_id)
          get_result_start = time.time()
          action, logits, value, rnn_state = agents[ep_id].result(response)
          get_result_time += time.time() - get_result_start
          take_action = agents[ep_id].take_action(action)
          action_dict[env_id][player_id] = take_action
          transition = Transition(
            step_type=timestep.step_type,
            input_dict=input_dict,
            action=action,
            rnn_state=rnn_state,
            reward=reward,
            policy_logits=logits,
            behaviour_value=value,
          )
          agents[ep_id].add(transition)
      envs_step_start = time.time()
      timesteps_dict = envs.step(action_dict)
      self._logger.write(
        {
          "time/get result": get_result_time,
          "time/envs step": time.time() - envs_step_start,
          "time/actor step": time.time() - actor_step_start,
        }
      )



"""Atati impala example."""
import os
from functools import partial
from typing import Any, Callable

import envpool
import launchpad as lp
from absl import app, flags, logging
from launchpad.nodes.dereference import Deferred
from launchpad.nodes.python.local_multi_processing import PythonProcess

from examples.atari.agent import AtariAgent
from examples.atari.network import network_maker
from moss.actor import GenericActor
from moss.buffer import QueueBuffer
from moss.env import EnvpoolVectorEnv
from moss.learner import ImpalaLearner
from moss.predictor import BasePredictor
from moss.types import Environment
from moss.utils.loggers import Logger, experiment_logger_factory
from moss.utils.paths import get_unique_id

flags.DEFINE_string("task_id", "Pong-v5", "Task name.")
flags.DEFINE_integer("stack_num", 1, "Stack nums.")
flags.DEFINE_integer("num_envs", 32, "Num of envs.")
flags.DEFINE_integer("num_threads", 6, "Num threads of envs.")
flags.DEFINE_bool(
  "use_resnet", False, "Use resnet or conv2d encoder for atari image."
)
flags.DEFINE_bool(
  "use_orthogonal", True, "Use orthogonal to initialization params weight."
)
flags.DEFINE_integer("num_actors", 16, "Num of actors.")
flags.DEFINE_integer("num_predictors", 4, "Num of predictors.")
flags.DEFINE_integer("unroll_len", 16, "Unroll length.")
flags.DEFINE_integer("predict_batch_size", 32, "Predict batch size.")
flags.DEFINE_integer("training_batch_size", 64, "Training batch size.")
flags.DEFINE_string("model_path", None, "Restore model path.")
flags.DEFINE_integer(
  "publish_interval", 1, "Publish params to predicotrs interval(by train steps)."
)
flags.DEFINE_integer("save_interval", 500, "Save interval(by train steps).")
flags.DEFINE_float("learning_rate", 5e-4, "Learning rate.")
flags.DEFINE_float("gamma", 0.99, "Reward discount rate.")
flags.DEFINE_float("gae_lambda", 0.95, "GAE lambda.")
flags.DEFINE_float("rho_clip", 0.9, "Clip threshold for importance ratios.")
flags.DEFINE_float(
  "pg_rho_clip", 0.9, "Clip threshold for policy gradient importance ratios."
)
flags.DEFINE_float("critic_coef", 0.25, "Critic coefficient.")
flags.DEFINE_float("entropy_coef", 0.01, "Entropy coefficient.")
flags.DEFINE_integer("buffer_size", 2048, "Replay buffer size.")
flags.DEFINE_string(
  "buffer_mode", "FIFO", "Queue buffer mode(Support `FIFO` and `LIFO`)."
)
flags.DEFINE_string("lp_launch_type", "local_mp", "Launch type.")

FLAGS = flags.FLAGS


def make_lp_program() -> Any:
  """Make launchpad program."""
  (exp_uid,) = get_unique_id()
  task_id = FLAGS.task_id
  stack_num = FLAGS.stack_num
  num_envs = FLAGS.num_envs
  num_threads = FLAGS.num_threads
  unroll_len = FLAGS.unroll_len

  dummy_env: Environment = envpool.make_dm(
    task_id, stack_num=stack_num, num_envs=1
  )
  obs_spec: Any = dummy_env.observation_spec()
  action_spec: Any = dummy_env.action_spec()
  use_resnet = FLAGS.use_resnet
  use_orthogonal = FLAGS.use_orthogonal

  logging.info(f"Task id: {task_id}")
  logging.info(f"Observation shape: {obs_spec.obs.shape}")
  logging.info(f"Action space: {action_spec.num_values}")

  def env_maker() -> EnvpoolVectorEnv:
    """Env maker."""
    return EnvpoolVectorEnv(
      task_id, stack_num=stack_num, num_envs=num_envs, num_threads=num_threads
    )

  def agent_maker(buffer: QueueBuffer, predictor: BasePredictor) -> Callable:
    """Agent maker."""

    def agent_wrapper(timestep: Any, logger: Logger) -> AtariAgent:
      """Return a agent."""
      del timestep
      return AtariAgent(unroll_len, buffer, predictor, logger)

    return agent_wrapper

  logger_fn = experiment_logger_factory(
    project=task_id, uid=exp_uid, time_delta=2.0, print_fn=logging.info
  )

  program = lp.Program("impala")
  with program.group("buffer"):
    buffer_node = lp.CourierNode(
      QueueBuffer,
      FLAGS.buffer_size,
      FLAGS.buffer_mode,
    )
    buffer = program.add_node(buffer_node)

  predictors = []
  with program.group("predictor"):
    for _ in range(FLAGS.num_predictors):
      predictor_node = lp.CourierNode(
        BasePredictor,
        FLAGS.predict_batch_size,
        partial(
          network_maker, obs_spec, action_spec, "NHWC", use_resnet,
          use_orthogonal
        ),
        logger_fn,
      )
      predictor = program.add_node(predictor_node)
      predictors.append(predictor)

  with program.group("actor"):
    for i in range(FLAGS.num_actors):
      actor_node = lp.CourierNode(
        GenericActor,
        Deferred(agent_maker, buffer, predictors[i % FLAGS.num_predictors]),
        env_maker,
        logger_fn,
      )
      program.add_node(actor_node)

  with program.group("learner"):
    save_path = os.path.join("checkpoints", task_id, exp_uid)
    learner_node = lp.CourierNode(
      ImpalaLearner,
      buffer,
      predictors,
      partial(
        network_maker, obs_spec, action_spec, "NHWC", use_resnet, use_orthogonal
      ),
      logger_fn,
      batch_size=FLAGS.training_batch_size,
      save_interval=FLAGS.save_interval,
      save_path=save_path,
      model_path=FLAGS.model_path,
      publish_interval=FLAGS.publish_interval,
      learning_rate=FLAGS.learning_rate,
      discount=FLAGS.gamma,
      gae_lambda=FLAGS.gae_lambda,
      clip_rho_threshold=FLAGS.rho_clip,
      clip_pg_rho_threshold=FLAGS.pg_rho_clip,
      critic_coef=FLAGS.critic_coef,
      entropy_coef=FLAGS.entropy_coef,
    )
    program.add_node(learner_node)
  return program


def main(_):
  """Main function."""
  program = make_lp_program()
  local_resources = {
    "actor":
      PythonProcess(
        env={
          "CUDA_VISIBLE_DEVICES": "",
          "XLA_PYTHON_CLIENT_PREALLOCATE": "false"
        }
      ),
    "learner":
      PythonProcess(
        env={
          "CUDA_VISIBLE_DEVICES": "0",
          "XLA_PYTHON_CLIENT_PREALLOCATE": "false"
        }
      ),
    "predictor":
      PythonProcess(
        env={
          "CUDA_VISIBLE_DEVICES": "0",
          "XLA_PYTHON_CLIENT_PREALLOCATE": "false"
        }
      ),
    "buffer":
      PythonProcess(
        env={
          "CUDA_VISIBLE_DEVICES": "",
          "XLA_PYTHON_CLIENT_PREALLOCATE": "false"
        }
      ),
  }
  lp.launch(program, local_resources=local_resources, terminal="tmux_session")


if __name__ == "__main__":
  os.environ["CUDA_VISIBLE_DEVICES"] = ""
  app.run(main)



"""Gym wrapper."""
from collections import namedtuple
from typing import Any, Dict, SupportsFloat, Tuple

import gymnasium as gym
from dm_env import StepType, TimeStep
from gymnasium.core import Env

from moss.types import Environment

Observation = namedtuple("Observation", ["obs", "info"])


class AutoResetWrapper(gym.Wrapper, gym.utils.RecordConstructorArgs):
  """Auto reset wrapper.

  The different with `gym.AutoResetWrapper` is `gym.AutoResetWrapper` ignore the
  last timestep, and put reward into next episode's first timestep.
  This wrapper follow envpool's auto reset. See more in:
  https://envpool.readthedocs.io/en/latest/content/python_interface.html#auto-reset
  """

  def __init__(self, env: Env) -> None:
    """Init."""
    gym.utils.RecordConstructorArgs.__init__(self)
    gym.Wrapper.__init__(self, env)
    self._need_reset: bool = False

  def step(self, action: Any) -> Tuple[Any, SupportsFloat, bool, bool, Dict]:
    """Env step."""
    if self._need_reset:
      self._need_reset = False
      obs, info = self.env.reset()
      reward, terminated, truncated = 0., False, False
      assert (
        "first_step" not in info
      ), "info dict cannot contain key 'first_step'."
      info["first_step"] = True
    else:
      obs, reward, terminated, truncated, info = self.env.step(  # type: ignore
        action
      )
      if terminated or truncated:
        self._need_reset = True
      assert (
        "first_step" not in info
      ), "info dict cannot contain key 'first_step'."
      info["first_step"] = False

    return obs, reward, terminated, truncated, info


class GymToDeepmindWrapper(Environment):
  """Wrapper `gym.Env` to `dm_env.Environment`."""

  def __init__(self, task_id: str, **kwargs: Any) -> None:
    """Init."""
    self._env = AutoResetWrapper(gym.make(task_id, **kwargs))

  def reset(self, **kwargs: Any) -> TimeStep:
    """Env reset."""
    obs, info = self._env.reset(**kwargs)
    timestep = TimeStep(
      step_type=StepType.FIRST,
      reward=0.,
      discount=1.,
      observation=Observation(obs, info),
    )
    return timestep

  def step(self, action: Any) -> TimeStep:
    """Env step."""
    obs, reward, terminated, truncated, info = self._env.step(action)
    if terminated:
      step_type = StepType.LAST
      discount = 0.0
    elif truncated:
      step_type = StepType.LAST
      discount = 1.0
    else:
      step_type = StepType.FIRST if info.get("first_step") else StepType.MID
      discount = 1.0
    timestep = TimeStep(
      step_type=step_type,
      reward=reward,
      discount=discount,
      observation=Observation(obs, info),
    )
    return timestep

  def action_spec(self) -> Any:
    """Get action spec."""
    return self._env.action_space

  def observation_spec(self) -> Any:
    """Get observation spec."""
    return self._env.observation_space

  def close(self) -> None:
    """Close env and release resources."""
    self._env.close()



"""Common network."""
from typing import Any, Callable, Dict, Optional, Tuple

import haiku as hk
import jax.numpy as jnp
import tree

from moss.network import Network
from moss.network.action import ActionSpec
from moss.network.feature import FeatureSpec
from moss.network.keys import AGENT_STATE, MASK
from moss.types import Array, KeyArray, NetOutput, Params, RNNState


class CommonModule(hk.Module):
  """Common haiku module."""

  def __init__(
    self,
    feature_spec: FeatureSpec,
    action_spec: ActionSpec,
    torso_net_maker: Callable[[], Any],
    value_net_maker: Callable[[], Any],
  ) -> None:
    """Init."""
    super().__init__("common_module")
    self._feature_spec = feature_spec
    self._feature_encoder = {
      name: (feature_set.process, feature_set.encoder_net_maker())
      for name, feature_set in feature_spec.feature_sets.items()
    }
    self._action_spec = action_spec
    self._torso_net = torso_net_maker()
    self._value_net = value_net_maker()

  def __call__(
    self, input_dict: Dict, rnn_state: RNNState, training: bool
  ) -> NetOutput:
    """Call."""
    embeddings = {}
    features = input_dict[AGENT_STATE]
    for name, feature_encoder in self._feature_encoder.items():
      feature = features[name]
      processor, encoder = feature_encoder
      if training:
        batch_encoder_apply = hk.BatchApply(encoder)
        batch_process_apply = hk.BatchApply(processor)
        embedding = batch_encoder_apply(batch_process_apply(feature))
      else:
        embedding = encoder(processor(feature))
      embeddings[name] = embedding

    if training:
      if isinstance(self._torso_net, hk.RNNCore):
        torso_out, rnn_state = hk.dynamic_unroll(
          self._torso_net, embeddings, rnn_state
        )
      else:
        batch_torso_apply = hk.BatchApply(self._torso_net)
        torso_out, rnn_state = batch_torso_apply(embeddings, rnn_state)
    else:
      torso_out, rnn_state = self._torso_net(embeddings, rnn_state)

    policy_logits = {}
    mask = input_dict.get(MASK, {})
    for name, action in self._action_spec.actions.items():
      action_mask = mask.get(action.name)
      if training:
        batch_policy_apply = hk.BatchApply(action.policy_net)
        policy_logits[name] = batch_policy_apply(torso_out, action_mask)
      else:
        policy_logits[name] = action.policy_net(torso_out, action_mask)

    if training:
      batch_value_apply = hk.BatchApply(self._value_net)
      value = batch_value_apply(torso_out)
    else:
      value = self._value_net(torso_out)

    return NetOutput(policy_logits, value, rnn_state)

  def initial_state(self, batch_size: Optional[int]) -> RNNState:
    """Constructs an initial state for rnn core."""
    try:
      rnn_state = self._torso_net.initial_state(batch_size)
    except Exception:
      rnn_state = None
    return rnn_state


class CommonNet(Network):
  """Common network."""

  def __init__(
    self,
    feature_spec: FeatureSpec,
    action_spec: ActionSpec,
    torso_net_maker: Callable[[], Any],
    value_net_maker: Callable[[], Any],
  ) -> None:
    """Init."""
    self._feature_spec = feature_spec
    self._action_spec = action_spec
    self._net = hk.without_apply_rng(
      hk.transform(
        lambda *args, **kwargs: CommonModule(
          feature_spec, action_spec, torso_net_maker, value_net_maker
        )(*args, **kwargs)
      )
    )
    self._initial_state = hk.without_apply_rng(
      hk.transform(
        lambda x: CommonModule(
          feature_spec, action_spec, torso_net_maker, value_net_maker
        ).initial_state(x)
      )
    )

  @property
  def action_spec(self) -> ActionSpec:
    """Action spec."""
    return self._action_spec

  def initial_params(self, rng: KeyArray) -> Params:
    """Init network's params."""
    dummy_agent_state = self._feature_spec.generate_value()
    dummy_inputs = {AGENT_STATE: dummy_agent_state}
    dummy_inputs = tree.map_structure(
      lambda x: jnp.expand_dims(x, 0), dummy_inputs
    )  # shape: [B, ...]
    init_rnn_state = self.initial_state(1)
    params = self._net.init(rng, dummy_inputs, init_rnn_state, False)
    return params

  def initial_state(self, batch_size: Optional[int]) -> RNNState:
    """Constructs an initial state for rnn core."""
    return self._initial_state.apply(None, batch_size)

  def forward(
    self, params: Params, input_dict: Dict, rnn_state: RNNState, rng: KeyArray,
    training: bool
  ) -> Tuple[Dict[str, Array], NetOutput]:
    """Network forward."""
    net_output = self._net.apply(params, input_dict, rnn_state, training)
    actions = self._action_spec.sample(rng, net_output.policy_logits)
    return actions, net_output



"""Atari utils."""
from typing import Any, List, Optional, Tuple

import envpool
import numpy as np
import pygame
import tree
from dm_env import Environment, TimeStep
from pygame import Surface


class LocalEnv(Environment):
  """Local environment wrapper for Atari games using EnvPool and Pygame.

  This class is a local environment wrapper that provides an interface
  to interact with Atari games. It is built on top of the EnvPool library,
  which provides efficient environment pooling for reinforcement learning,
  and uses Pygame for rendering the game frames.

  Attributes:
    task_id (str): Identifier for the specific Atari game.
    env (Environment): The EnvPool environment instance for the game logic.
    render (bool): Flag to determine whether to render the game frames.
    fps (int): Frames per second for game rendering.
    scale (int): Scaling factor for rendering the game frames.
    window (Optional[Surface]): Pygame window surface for rendering.
    screen_size (Optional[Tuple[float, float]]): The size of the rendering
      window.
    clock: Pygame clock object for controlling the render loop timing.

  Args:
    task_id (str): Identifier for the specific Atari game.
    seed (int, optional): Random seed for the environment.
    render (bool, optional): Whether to render the game frames.
    fps (int, optional): Frames per second for game rendering.
    scale (int, optional): Scaling factor for rendering the game frames.

  """

  def __init__(
    self,
    task_id: str,
    seed: int = 42,
    render: bool = True,
    fps: int = 60,
    scale: int = 1
  ) -> None:
    """Initializes the local Atari environment with optional rendering.

    Args:
      task_id (str): Identifier for the specific Atari game.
      seed (int): Random seed for the environment.
      render (bool): Whether to render the game frames.
      fps (int): Frames per second for game rendering.
      scale (int): Scaling factor for rendering the game frames.
    """
    self._task_id = task_id
    self._env: Environment = envpool.make_dm(
      task_id, seed=seed, stack_num=1, num_envs=1
    )
    self._render_mode = render
    if render:
      # set the same seed to env and render env.
      self._render_env: Environment = envpool.make_dm(
        task_id,
        seed=seed,
        stack_num=1,
        num_envs=1,
        img_height=210,
        img_width=160,
        gray_scale=False,
      )
      self._fps = fps
      self._scale = scale
      self._window: Optional[Surface] = None
      self._screen_size: Optional[Tuple[float, float]] = None
      self._clock = pygame.time.Clock()

  def _split_batch_timestep(self, batch: TimeStep) -> List[TimeStep]:
    """Splits a batched timestep into a list of individual timesteps.

    Args:
      batch (TimeStep): The batched timestep to split.

    Returns:
      List[TimeStep]: A list of individual timesteps.
    """
    size = batch.step_type.size
    timesteps = [
      tree.map_structure(lambda x: x[i], batch)  # noqa: B023
      for i in range(size)
    ]
    return timesteps

  def reset(self) -> TimeStep:
    """Resets the environment and returns the initial observation.

    Returns:
        TimeStep: The initial timestep after resetting the environment.
    """
    if self._render_mode:
      render_timestep = self._render_env.reset()
      render_timestep = self._split_batch_timestep(render_timestep)[0]
      self.render(render_timestep)
    timestep = self._env.reset()
    timestep = self._split_batch_timestep(timestep)[0]
    return timestep

  def step(self, action) -> TimeStep:
    """Steps the environment and returns the new timestep.

    Args:
      action: The action to take in the environment.

    Returns:
      TimeStep: The timestep resulting from the action.
    """
    if self._render_mode:
      render_timestep = self._render_env.step(action)
      render_timestep = self._split_batch_timestep(render_timestep)[0]
      self.render(render_timestep)
    timestep = self._env.step(action)
    timestep = self._split_batch_timestep(timestep)[0]
    return timestep

  def observation_spec(self) -> Any:
    """Defines the observations provided by the environment.

    May use a subclass of `specs.Array` that specifies additional properties
    such as min and max bounds on the values.

    Returns:
      An `Array` spec, or a nested dict, list or tuple of `Array` specs.
    """
    return self._env.observation_spec()

  def action_spec(self) -> Any:
    """Defines the actions that should be provided to `step`.

    May use a subclass of `specs.Array` that specifies additional properties
    such as min and max bounds on the values.

    Returns:
      An `Array` spec, or a nested dict, list or tuple of `Array` specs.
    """
    return self._env.action_spec()

  def render(self, timestep: TimeStep) -> None:
    """Renders the current game frame based on the latest timestep.

    Args:
      timestep (TimeStep): The timestep to render.
    """
    obs = timestep.observation.obs  # type: ignore
    rgb_array = np.transpose(obs, axes=(2, 1, 0))  # (height, width, channel)

    if self._screen_size is None:
      width, height = rgb_array.shape[:2]
      self._screen_size = (width * self._scale, height * self._scale)

    if self._window is None:
      pygame.init()
      pygame.display.init()
      pygame.display.set_caption(self._task_id)
      self._window = pygame.display.set_mode(self._screen_size)

    surf = pygame.surfarray.make_surface(rgb_array)
    surf = pygame.transform.scale(surf, self._screen_size)
    self._window.blit(surf, (0, 0))
    pygame.event.pump()
    self._clock.tick(self._fps)
    pygame.display.flip()

  def close(self) -> None:
    """Releases resources and closes rendering window."""
    self._env.close()
    if self._render_mode:
      self._render_env.close()
      if self._window is not None:
        pygame.display.quit()
        pygame.quit()



"""Impala learner."""
from functools import partial
from typing import Callable, Dict, List, Optional, Tuple

import chex
import distrax
import jax
import jax.numpy as jnp
import rlax

from moss.core import Buffer, Predictor
from moss.learner.base import BaseLearner
from moss.network import Network
from moss.types import Array, LoggingData, Params, StepType, Transition
from moss.utils.loggers import Logger


class ImpalaLearner(BaseLearner):
  """Impala learner."""

  def __init__(
    self,
    buffer: Buffer,
    predictors: List[Predictor],
    network_maker: Callable[[], Network],
    logger_fn: Callable[..., Logger],
    batch_size: int,
    save_interval: int,
    save_path: str,
    model_path: Optional[str] = None,
    gradient_clip: Optional[float] = None,
    data_reuse: Optional[int] = None,
    publish_interval: int = 1,
    learning_rate: float = 5e-4,
    discount: float = 0.99,
    gae_lambda: float = 0.95,
    clip_rho_threshold: float = 1.0,
    clip_pg_rho_threshold: float = 1.0,
    critic_coef: float = 0.5,
    entropy_coef: float = 0.01,
    seed: int = 42,
  ) -> None:
    """Init."""
    super().__init__(
      buffer, predictors, network_maker, logger_fn, batch_size, save_interval,
      save_path, model_path, gradient_clip, data_reuse, publish_interval,
      learning_rate, seed
    )
    self._discount = discount
    self._gae_lambda = gae_lambda
    self._clip_rho_threshold = clip_rho_threshold
    self._clip_pg_rho_threshold = clip_pg_rho_threshold
    self._critic_coef = critic_coef
    self._entropy_coef = entropy_coef

  def _policy_gradient_loss(
    self,
    logits_t: Dict[str, Array],
    a_t: Dict[str, Array],
    adv_t: Array,
    w_t: Array,
    use_stop_gradient: bool = True,
  ) -> Array:
    """Calculates the policy gradient loss.

    See "Simple Gradient-Following Algorithms for Connectionist RL" by Williams.
    (http://www-anw.cs.umass.edu/~barto/courses/cs687/williams92simple.pdf)

    Args:
      a_t: a sequence of actions sampled from the preferences `logits_t`.
      adv_t: the observed or estimated advantages from executing actions `a_t`.
      w_t: a per timestep weighting for the loss.
      use_stop_gradient: bool indicating whether or not to apply stop gradient to
        advantages.

    Returns:
      Loss whose gradient corresponds to a policy gradient update.
    """
    for logits in logits_t.values():
      chex.assert_rank(logits, 2)
      chex.assert_type(logits, float)
    for action in a_t.values():
      chex.assert_rank(action, 1)
      chex.assert_type(action, int)
    chex.assert_rank([adv_t, w_t], [1, 1])
    chex.assert_type([adv_t, w_t], [float, float])

    distribution = self._network.action_spec.distribution(logits_t)
    log_pi_a_t = distribution.log_prob(a_t)
    adv_t = jax.lax.select(
      use_stop_gradient, jax.lax.stop_gradient(adv_t), adv_t
    )
    loss_per_timestep = -log_pi_a_t * adv_t
    return jnp.mean(loss_per_timestep * w_t)

  def _entropy_loss(
    self,
    logits_t: Dict[str, Array],
    w_t: Array,
  ) -> Array:
    """Calculates the entropy regularization loss.

    See "Function Optimization using Connectionist RL Algorithms" by Williams.
    (https://www.tandfonline.com/doi/abs/10.1080/09540099108946587)

    Args:
      logits_t: a sequence of unnormalized action preferences.
      w_t: a per timestep weighting for the loss.

    Returns:
      Entropy loss.
    """
    for logits in logits_t.values():
      chex.assert_rank(logits, 2)
      chex.assert_type(logits, float)
    chex.assert_rank(w_t, 1)
    chex.assert_type(w_t, float)

    distribution = self._network.action_spec.distribution(logits_t)
    entropy_per_timestep = distribution.entropy()
    return -jnp.mean(entropy_per_timestep * w_t)

  def _loss(self, params: Params, data: Transition) -> Tuple[Array, LoggingData]:
    """Impala loss."""
    rnn_state = jax.tree_map(lambda x: x[0], data.rnn_state)
    _, net_output = self._network.forward(
      params, data.input_dict, rnn_state, jax.random.PRNGKey(0), True
    )

    actions, rewards = data.action, data.reward
    behaviour_logits = data.policy_logits
    learner_logits, values = net_output.policy_logits, net_output.value
    discount = jnp.ones_like(data.step_type) * self._discount
    # The step is uninteresting if we transitioned LAST -> FIRST.
    mask = jnp.not_equal(data.step_type[:-1], int(StepType.FIRST))
    mask = mask.astype(jnp.float32)

    actions_tm1 = jax.tree_map(lambda x: x[:-1], actions)
    behaviour_logits_tm1 = jax.tree_map(lambda x: x[:-1], behaviour_logits)
    learner_logits_tm1 = jax.tree_map(lambda x: x[:-1], learner_logits)
    values_tm1, values_t = values[:-1], values[1:]
    rewards_t, discount_t = rewards[1:], discount[1:]

    # Importance sampling.
    rhos = distrax.importance_sampling_ratios(
      self._network.action_spec.distribution(learner_logits_tm1),
      self._network.action_spec.distribution(behaviour_logits_tm1),
      actions_tm1,
    )

    # Critic loss.
    vtrace_td_error_and_advantage_fn = partial(
      rlax.vtrace_td_error_and_advantage,
      lambda_=self._gae_lambda,
      clip_rho_threshold=self._clip_rho_threshold,
      clip_pg_rho_threshold=self._clip_pg_rho_threshold
    )
    vmap_vtrace_td_error_and_advantage_fn = jax.vmap(
      vtrace_td_error_and_advantage_fn, in_axes=1, out_axes=1
    )
    vtrace_returns = vmap_vtrace_td_error_and_advantage_fn(
      values_tm1, values_t, rewards_t, discount_t, rhos
    )
    critic_loss = jnp.mean(jnp.square(vtrace_returns.errors) * mask)
    critic_loss = self._critic_coef * critic_loss

    # Policy gradien loss.
    vmap_policy_gradient_loss_fn = jax.vmap(
      self._policy_gradient_loss, in_axes=1, out_axes=0
    )
    pg_loss = vmap_policy_gradient_loss_fn(
      learner_logits_tm1, actions_tm1, vtrace_returns.pg_advantage, mask
    )
    pg_loss = jnp.mean(pg_loss)

    # Entropy loss.
    vmap_entropy_loss_fn = jax.vmap(self._entropy_loss, in_axes=1, out_axes=0)
    entropy_loss = vmap_entropy_loss_fn(learner_logits_tm1, mask)
    entropy_loss = self._entropy_coef * jnp.mean(entropy_loss)

    # Total loss.
    total_loss = pg_loss + critic_loss + entropy_loss

    # Metrics.
    metrics = {
      "loss/policy": pg_loss,
      "loss/critic": critic_loss,
      "loss/entropy": entropy_loss,
      "loss/total": total_loss,
    }
    return total_loss, metrics



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Loggers which filter other loggers."""

import math
import time
from typing import Callable, Optional, Sequence

from moss.utils.loggers import base


class NoneFilter(base.Logger):
  """Logger which writes to another logger, filtering any `None` values."""

  def __init__(self, to: base.Logger):
    """Initializes the logger.

    Args:
      to: A `Logger` object to which the current object will forward its results
        when `write` is called.
    """
    self._to = to

  def write(self, values: base.LoggingData) -> None:
    """Write."""
    values = {k: v for k, v in values.items() if v is not None}
    self._to.write(values)

  def close(self) -> None:
    """Close."""
    self._to.close()


class TimeFilter(base.Logger):
  """Logger which writes to another logger at a given time interval."""

  def __init__(self, to: base.Logger, time_delta: float):
    """Initializes the logger.

    Args:
      to: A `Logger` object to which the current object will forward its results
        when `write` is called.
      time_delta: How often to write values out in seconds.
        Note that writes within `time_delta` are dropped.
    """
    self._to = to
    self._time = 0
    self._time_delta = time_delta
    if time_delta < 0:
      raise ValueError(f'time_delta must be greater than 0 (got {time_delta}).')

  def write(self, values: base.LoggingData) -> None:
    """Write."""
    now = time.time()
    if (now - self._time) > self._time_delta:
      self._to.write(values)
      self._time = now  # type: ignore

  def close(self) -> None:
    """Close."""
    self._to.close()


class KeyFilter(base.Logger):
  """Logger which filters keys in logged data."""

  def __init__(
    self,
    to: base.Logger,
    *,
    keep: Optional[Sequence[str]] = None,
    drop: Optional[Sequence[str]] = None,
  ):
    """Creates the filter.

    Args:
      to: A `Logger` object to which the current object will forward its writes.
      keep: Keys that are kept by the filter. Note that `keep` and `drop` cannot
        be both set at once.
      drop: Keys that are dropped by the filter. Note that `keep` and `drop`
        cannot be both set at once.
    """
    if bool(keep) == bool(drop):
      raise ValueError('Exactly one of `keep` & `drop` arguments must be set.')
    self._to = to
    self._keep = keep
    self._drop = drop

  def write(self, data: base.LoggingData) -> None:
    """Write."""
    if self._keep:
      data = {k: data[k] for k in self._keep}
    if self._drop:
      data = {k: v for k, v in data.items() if k not in self._drop}
    self._to.write(data)

  def close(self) -> None:
    """Close."""
    self._to.close()


class GatedFilter(base.Logger):
  """Logger which writes to another logger based on a gating function.

  This logger tracks the number of times its `write` method is called, and uses
  a gating function on this number to decide when to write.
  """

  def __init__(self, to: base.Logger, gating_fn: Callable[[int], bool]):
    """Initialises the logger.

    Args:
      to: A `Logger` object to which the current object will forward its results
        when `write` is called.
      gating_fn: A function that takes an integer (number of calls) as input.
        For example, to log every tenth call: gating_fn=lambda t: t % 10 == 0.
    """
    self._to = to
    self._gating_fn = gating_fn
    self._calls = 0

  def write(self, values: base.LoggingData) -> None:
    """Write."""
    if self._gating_fn(self._calls):
      self._to.write(values)
    self._calls += 1

  def close(self) -> None:
    """Close."""
    self._to.close()

  @classmethod
  def logarithmic(cls, to: base.Logger, n: int = 10) -> 'GatedFilter':
    """Builds a logger for writing at logarithmically-spaced intervals.

    This will log on a linear scale at each order of magnitude of `n`.
    For example, with n=10, this will log at times:
        [0, 1, 2, ..., 9, 10, 20, 30, ... 90, 100, 200, 300, ... 900, 1000]

    Args:
      to: The underlying logger to write to.
      n: Base (default 10) on which to operate.

    Returns:
      A GatedFilter logger, which gates logarithmically as described above.
    """

    def logarithmic_filter(t: int) -> bool:
      magnitude = math.floor(math.log10(max(t, 1)) / math.log10(n))
      return t % (n**magnitude) == 0

    return cls(to, gating_fn=logarithmic_filter)

  @classmethod
  def periodic(cls, to: base.Logger, interval: int = 10) -> 'GatedFilter':
    """Builds a logger for writing at linearly-spaced intervals.

    Args:
      to: The underlying logger to write to.
      interval: The interval between writes.

    Returns:
      A GatedFilter logger, which gates periodically as described above.
    """
    return cls(to, gating_fn=lambda t: t % interval == 0)



"""PPO learner."""
from functools import partial
from typing import Callable, Dict, List, Optional, Tuple

import chex
import distrax
import jax
import jax.numpy as jnp
import rlax

from moss.core import Buffer, Predictor
from moss.learner.base import BaseLearner
from moss.network import Network
from moss.types import Array, LoggingData, Params, StepType, Transition
from moss.utils.loggers import Logger


class PPOLearner(BaseLearner):
  """PPO learner."""

  def __init__(
    self,
    buffer: Buffer,
    predictors: List[Predictor],
    network_maker: Callable[[], Network],
    logger_fn: Callable[..., Logger],
    batch_size: int,
    save_interval: int,
    save_path: str,
    model_path: Optional[str] = None,
    gradient_clip: Optional[float] = None,
    data_reuse: Optional[int] = None,
    publish_interval: int = 1,
    learning_rate: float = 5e-4,
    discount: float = 0.99,
    gae_lambda: float = 0.95,
    pg_clip_epsilon: float = 0.1,
    value_clip_epsilon: Optional[float] = None,
    critic_coef: float = 0.5,
    entropy_coef: float = 0.01,
    seed: int = 42,
  ) -> None:
    """Init."""
    super().__init__(
      buffer, predictors, network_maker, logger_fn, batch_size, save_interval,
      save_path, model_path, gradient_clip, data_reuse, publish_interval,
      learning_rate, seed
    )
    self._discount = discount
    self._gae_lambda = gae_lambda
    self._pg_clip_epsilon = pg_clip_epsilon
    self._value_clip_epsilon = value_clip_epsilon
    self._critic_coef = critic_coef
    self._entropy_coef = entropy_coef

  def _entropy_loss(
    self,
    logits_t: Dict[str, Array],
    w_t: Array,
  ) -> Array:
    """Calculates the entropy regularization loss.

    See "Function Optimization using Connectionist RL Algorithms" by Williams.
    (https://www.tandfonline.com/doi/abs/10.1080/09540099108946587)

    Args:
      logits_t: a sequence of unnormalized action preferences.
      w_t: a per timestep weighting for the loss.

    Returns:
      Entropy loss.
    """
    for logits in logits_t.values():
      chex.assert_rank(logits, 2)
      chex.assert_type(logits, float)
    chex.assert_rank(w_t, 1)
    chex.assert_type(w_t, float)

    distribution = self._network.action_spec.distribution(logits_t)
    entropy_per_timestep = distribution.entropy()
    return -jnp.mean(entropy_per_timestep * w_t)

  def _loss(self, params: Params, data: Transition) -> Tuple[Array, LoggingData]:
    """PPO loss."""
    rnn_state = jax.tree_map(lambda x: x[0], data.rnn_state)
    _, net_output = self._network.forward(
      params, data.input_dict, rnn_state, jax.random.PRNGKey(0), True
    )

    actions, rewards = data.action, data.reward
    behaviour_logits = data.policy_logits
    behaviour_values = data.behaviour_value
    learner_logits, values = net_output.policy_logits, net_output.value
    discount = jnp.ones_like(data.step_type) * self._discount
    # The step is uninteresting if we transitioned LAST -> FIRST.
    mask = jnp.not_equal(data.step_type[:-1], int(StepType.FIRST))
    mask = mask.astype(jnp.float32)

    actions_tm1 = jax.tree_map(lambda x: x[:-1], actions)
    behaviour_logits_tm1 = jax.tree_map(lambda x: x[:-1], behaviour_logits)
    learner_logits_tm1 = jax.tree_map(lambda x: x[:-1], learner_logits)
    values_tm1, behavior_values_tm1 = values[:-1], behaviour_values[:-1]
    rewards_t, discount_t = rewards[1:], discount[1:]

    # Importance sampling.
    rhos = distrax.importance_sampling_ratios(
      self._network.action_spec.distribution(learner_logits_tm1),
      self._network.action_spec.distribution(behaviour_logits_tm1),
      actions_tm1,
    )

    # Computes GAE.
    vmap_generalized_advantage_estimation_fn = jax.vmap(
      rlax.truncated_generalized_advantage_estimation,
      in_axes=[1, 1, None, 1],
      out_axes=1
    )
    advantage_tm1 = vmap_generalized_advantage_estimation_fn(
      rewards_t, discount_t, self._gae_lambda, values
    )

    # Policy gradient loss.
    clipped_surrogate_pg_loss_fn = partial(
      rlax.clipped_surrogate_pg_loss, epsilon=self._pg_clip_epsilon
    )
    vmap_clipped_surrogate_pg_loss_fn = jax.vmap(
      clipped_surrogate_pg_loss_fn, in_axes=1, out_axes=0
    )
    clipped_surrogate_pg_loss = vmap_clipped_surrogate_pg_loss_fn(
      rhos, advantage_tm1
    )
    pg_loss = jnp.mean(clipped_surrogate_pg_loss)

    # Critic loss.
    td_targets = jax.lax.stop_gradient(advantage_tm1 + values_tm1)
    unclipped_td_errors = td_targets - values_tm1
    unclipped_critic_loss = jnp.square(unclipped_td_errors)
    if self._value_clip_epsilon is not None:
      # Clip values to reduce variablility during critic training.
      clipped_values_tm1 = behavior_values_tm1 + jnp.clip(
        values_tm1 - behavior_values_tm1,
        -self._value_clip_epsilon,
        self._value_clip_epsilon,
      )
      clipped_td_errors = td_targets - clipped_values_tm1
      clipped_critic_loss = jnp.square(clipped_td_errors)
      critic_loss = jnp.mean(
        jnp.fmax(unclipped_critic_loss, clipped_critic_loss) * mask
      )
    else:
      critic_loss = jnp.mean(unclipped_critic_loss * mask)
    critic_loss = self._critic_coef * critic_loss

    # Entropy loss.
    vmap_entropy_loss_fn = jax.vmap(self._entropy_loss, in_axes=1, out_axes=0)
    entropy_loss = vmap_entropy_loss_fn(learner_logits_tm1, mask)
    entropy_loss = self._entropy_coef * jnp.mean(entropy_loss)

    # Total loss.
    total_loss = pg_loss + critic_loss + entropy_loss

    # Metrics.
    metrics = {
      "loss/policy": pg_loss,
      "loss/critic": critic_loss,
      "loss/entropy": entropy_loss,
      "loss/total": total_loss,
    }
    return total_loss, metrics



"""Base environment worker."""
import abc
from typing import Any, Callable, Dict

from moss.env.base import BaseEnv
from moss.types import TimeStep

try:
  import ray
except ImportError:
  ray = None  # type: ignore

AgentID = Any


class BaseEnvWorker(abc.ABC):
  """Base environment worker."""

  def __init__(self, env_maker: Callable[[], BaseEnv], **kwargs: Any) -> None:
    """Init."""
    self._env = env_maker(**kwargs)

  @abc.abstractmethod
  def reset(self) -> Dict[AgentID, TimeStep]:
    """Reset."""

  @abc.abstractmethod
  def step(self, actions: Any) -> Dict[AgentID, TimeStep]:
    """Step."""

  @property
  def env(self) -> BaseEnv:
    """Get env."""
    return self._env


class DummyWorker(BaseEnvWorker):
  """Dummy environment worker."""

  def __init__(self, env_maker: Callable[[], BaseEnv], **kwargs: Any) -> None:
    """Init."""
    super().__init__(env_maker, **kwargs)

  def reset(self) -> Any:
    """Dummy worker reset."""
    return self._env.reset()

  def step(self, actions: Any) -> Any:
    """Dummy worker step."""
    return self._env.step(actions)


class RayEnvWorker(BaseEnvWorker):
  """Ray env worker."""

  def __init__(self, env_maker: Callable[[], BaseEnv], **kwargs: Any) -> None:
    """Init."""
    if ray is None:
      raise ImportError(
        "Please install ray to support RayVectorEnv: pip install ray"
      )
    self._env = ray.remote(num_cpus=0)(DummyWorker).remote(
      env_maker, **kwargs
    )  # type: ignore

  def reset(self) -> Any:
    """Call ray env worker reset remote."""
    return ray.get(self._env.reset.remote())  # type: ignore

  def step(self, actions: Any) -> Any:
    """Call ray env step remote."""
    return ray.get(self._env.step.remote(actions))  # type: ignore



"""PettingZoo environment suite moss."""
from collections import namedtuple
from typing import Any, Dict, List

from pettingzoo.utils.all_modules import all_environments

from moss.env.base import BaseEnv
from moss.env.wrapper import PettingZooToDeepmindWrapper
from moss.types import TimeStep

AgentID = Any
Observation = namedtuple("Observation", ["obs", "info"])

all_envs = {**all_environments}
for env_id, env_module in all_environments.items():
  all_envs[env_id.split("/")[-1]] = env_module


def make_pettingzoo(
  task_id: str,
  height: int = 84,
  width: int = 84,
  **kwargs: Any
) -> PettingZooToDeepmindWrapper:
  """Pettingzoo env maker."""
  task_id = task_id.replace("-", "_").lower()
  if task_id not in all_envs:
    raise ValueError(f"Not support env `{task_id}`, please check you env name.")
  env_module = all_envs[task_id]
  return PettingZooToDeepmindWrapper(
    env_module.parallel_env, height=height, width=width, **kwargs
  )


class PettingZooEnv(BaseEnv):
  """PetingZoo env."""

  def __init__(
    self,
    task_id: str,
    height: int = 84,
    width: int = 84,
    **kwargs: Any
  ) -> None:
    """Init."""
    self._env = make_pettingzoo(task_id, height, width, **kwargs)

  def reset(self) -> Dict[AgentID, TimeStep]:
    """Reset."""
    timestep = self._env.reset()
    return self.timestep_process(timestep)

  def step(self, action: Any) -> Dict[AgentID, TimeStep]:
    """Step."""
    action = self.action_process(action)
    timestep = self._env.step(action)
    return self.timestep_process(timestep)

  def timestep_process(self, timestep: TimeStep) -> Dict[AgentID, TimeStep]:
    """Timestep process."""
    new_timesteps = {}
    for agent_id in self.agents:
      new_timesteps[agent_id] = TimeStep(
        step_type=timestep.step_type[agent_id],
        reward=timestep.reward[agent_id],
        discount=timestep.discount[agent_id],
        observation=Observation(
          timestep.observation.obs[agent_id],
          timestep.observation.info[agent_id],
        ),
      )
    return new_timesteps

  def action_process(self, action: Dict[AgentID, Any]) -> Dict[AgentID, Any]:
    """Action process."""
    return action

  @property
  def agents(self) -> List[AgentID]:
    """Get all agent id."""
    return self._env.agents

  def action_spec(self) -> Any:
    """Get agent action spec by agent id."""
    return self._env.action_spec()

  def observation_spec(self) -> Any:
    """Get agent observation spec by agent id."""
    return self._env.observation_spec()

  def close(self) -> None:
    """Close env and release resources."""
    self._env.close()



import os
from typing import Dict, List

from setuptools import find_packages, setup


def get_version() -> str:
  # https://packaging.python.org/guides/single-sourcing-package-version/
  init = open(os.path.join("moss", "__init__.py"), "r").read().split()
  return init[init.index("__version__") + 2][1:-1]


def get_install_requires() -> List[str]:
  launchpad = [
    "tensorflow==2.8.1",
    "dm-reverb==0.12.0",
    "dm-launchpad==0.5.2",
  ]

  jax_requirements = [
    "jax==0.4.9",
    "jaxlib==0.4.9",
    "dm-haiku==0.0.10",
    "chex",
    "flax",
    "optax",
    "rlax",
  ] + launchpad

  requirements = [
    "absl-py",
    "envpool",
    "dm-env",
    "dm-tree",
    "numpy",
    "pillow",
    "pygame",
    "tensorboardX",
    "tqdm",
    "typing-extensions",
  ] + jax_requirements
  return requirements


def get_extras_require() -> Dict[str, List[str]]:
  req = {
    "dev":
      [
        "sphinx<7",
        "sphinx_rtd_theme",
        "jinja2",
        "sphinxcontrib-bibtex",
        "flake8",
        "flake8-bugbear",
        "yapf",
        "isort",
        "mypy",
        "pydocstyle",
        "doc8",
        "pyenchant",
        "pytest",
        "pytest-cov",
      ],
  }
  return req


setup(
  name="moss-rl",
  version=get_version(),
  description="A Python library for Reinforcement Learning.",
  long_description=open("README.md", encoding="utf8").read(),
  long_description_content_type="text/markdown",
  author="lanzhiyi",
  author_email="lanzhiy.mail@qq.com",
  url="https://github.com/hilanzy/moss",
  packages=find_packages(),
  license="MIT",
  python_requires=">=3.8",
  install_requires=get_install_requires(),
  extras_require=get_extras_require(),
  classifiers=[
    # How mature is this project? Common values are
    #   3 - Alpha
    #   4 - Beta
    #   5 - Production/Stable
    "Development Status :: 3 - Alpha",
    # Indicate who your project is intended for
    "Intended Audience :: Developers",
    "Topic :: Software Development :: Libraries :: Python Modules",
    # Pick your license as you wish (should match "license" above)
    "License :: OSI Approved :: MIT License",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
  ],
  keywords="reinforcement-learning python machine learning",
)



"""Vizdoom example utils."""
import envpool
from dm_env import Environment


def vizdoom_env_maker(
  map_id: str, num_envs: int, cfg_path: str, wad_path: str, **kwargs
) -> Environment:
  """Vizdoom env maker."""
  task_id = "".join([i.capitalize() for i in map_id.split("_")]) + "-v1"
  reward_config = {
    "KILLCOUNT": [20.0, -20.0],
    "HEALTH": [1.0, 0.0],
    "AMMO2": [1.0, -1.0],
  }
  if "battle" in task_id:
    reward_config["HEALTH"] = [1.0, -1.0]
  return envpool.make_dm(
    task_id,
    num_envs=num_envs,
    cfg_path=cfg_path,
    wad_path=wad_path,
    reward_config=reward_config,
    use_combined_action=True,
    max_episode_steps=2625,
    use_inter_area_resize=False,
    **kwargs,
  )



"""Base learner."""
import abc
import os
import pickle
import time
from functools import partial
from typing import Callable, List, Optional, Tuple

import jax
import optax
from absl import logging

from moss.core import Buffer, Learner, Predictor
from moss.network import Network
from moss.types import Array, LoggingData, OptState, Params, Transition
from moss.utils.loggers import Logger


class BaseLearner(Learner):
  """Base learner."""

  def __init__(
    self,
    buffer: Buffer,
    predictors: List[Predictor],
    network_maker: Callable[[], Network],
    logger_fn: Callable[..., Logger],
    batch_size: int,
    save_interval: int,
    save_path: str,
    model_path: Optional[str] = None,
    gradient_clip: Optional[float] = None,
    data_reuse: Optional[int] = None,
    publish_interval: int = 1,
    learning_rate: float = 5e-4,
    seed: int = 42,
  ) -> None:
    """Init."""
    self._buffer = buffer
    self._predictors = predictors
    self._network = network_maker()
    self._batch_size = batch_size
    self._logger = logger_fn(label="Learner")
    if model_path is not None:
      self._params = self._load_model(model_path)
    else:
      self._params = self._initial_params(seed)
    self._publish_params(self._params)
    self._optmizer = optax.rmsprop(learning_rate, decay=0.99, eps=1e-7)
    if gradient_clip is not None:
      self._optmizer = optax.chain(
        optax.clip_by_global_norm(gradient_clip), self._optmizer
      )
    self._opt_state = self._optmizer.init(self._params)
    self._save_intelval = save_interval
    self._save_fn = partial(self._save_model, save_path=save_path)
    self._data_reuse = data_reuse or 1
    self._publish_interval = publish_interval
    logging.info(jax.devices())

  def _initial_params(self, seed: int) -> Params:
    """Init params and update to predictor."""
    rng = jax.random.PRNGKey(seed)
    params = self._network.initial_params(rng)
    return params

  def _publish_params(self, params: Params) -> None:
    """Publish params to predictor."""
    for predictor in self._predictors:
      predictor.update_params(params)

  @abc.abstractmethod
  def _loss(self, params: Params, data: Transition) -> Tuple[Array, LoggingData]:
    """Loss function."""

  @partial(jax.jit, static_argnums=0)
  def _train_step(self, params: Params, opt_state: OptState,
                  data: Transition) -> Tuple[Params, OptState, LoggingData]:
    """Training step."""
    (_, metrics), grads = jax.value_and_grad(
      self._loss, has_aux=True
    )(params, data)
    updates, opt_state = self._optmizer.update(grads, opt_state)
    params = optax.apply_updates(params, updates)
    return params, opt_state, metrics

  def _load_model(self, model_path: str) -> Params:
    """Load model."""
    with open(model_path, mode="rb") as f:
      params: Params = pickle.load(f)
    return params

  def _save_model(self, save_path: str, params: Params) -> None:
    """Save model."""
    os.makedirs(save_path, exist_ok=True)
    suffix = int(time.time())
    save_file = os.path.join(save_path, f"model-{suffix}.pkl")
    params_pkl = pickle.dumps(params)
    with open(save_file, mode="wb") as f:
      f.write(params_pkl)

  def run(self) -> None:
    """Run learner."""
    train_steps = 0
    while True:
      logs = {}
      start_sample_time = time.time()
      training_data = self._buffer.sample(self._batch_size)
      sample_data_time = time.time() - start_sample_time

      start_training_time = time.time()
      for _ in range(self._data_reuse):
        self._params, self._opt_state, metrics = self._train_step(
          self._params, self._opt_state, training_data
        )
        train_steps += 1
        self._logger.write(metrics)
      training_step_time = (time.time() - start_training_time) / self._data_reuse

      publish_params_start = time.time()
      if train_steps % self._publish_interval == 0:
        self._publish_params(self._params)
      publish_params_time = time.time() - publish_params_start

      if train_steps % self._save_intelval == 0:
        self._save_fn(params=self._params)

      logs.update(
        {
          "time/sample data": sample_data_time,
          "time/training step": training_step_time,
          "time/publish params": publish_params_time,
          "train steps": train_steps,
        }
      )
      self._logger.write(logs)
